#!/usr/bin/env python
# -*- coding: utf-8 -*-
import tkinter as tk
from tkinter import ttk
import time
from gui import gui_scrolltext
from gui import gui_script_handling
import options as options_module
import numpy as np
font='Arial 9 '
tags=dict()
tags['option']={'fg':'#025714','bg':'#e6eaf0','font':font+'bold'}
tags['unselected']={'fg':'black','bg':'white','font':font}

class optionset(dict):
	def __init__(self,tabs,window):
		dict.__init__(self)
		#s = ttk.Style()
		#s.configure('new.TFrame', background='white',font=font)			
		self.tabs=tabs
		self.tabs.rowconfigure(0,weight=1)
		self.tabs.columnconfigure(0,weight=1)		
		self.win=window
		self.default_options=options_module.regression_options()
		self.main_frame=tk.Frame(tabs,background='white')	
		self.main_frame.rowconfigure(0,weight=1)
		self.main_frame.columnconfigure(0,weight=1)			
		self.main_frame.grid(row=0,column=0,sticky=tk.NSEW)
		self.frames=dict()
		self.frames['default']=tk.Frame(self.main_frame,background='white')
		self.default_msg=tk.Label(self.frames['default'],text='Please select a data set before editing options',background='white')
		self.frames['default'].grid()
		self.option_frame=options_item(self.frames['default'], self.win,self.default_options,'Options')
		self.option_frame.register_validation()


		
def add_preferences_tab(tabs,window):
	tabs.rowconfigure(0,weight=1)
	tabs.columnconfigure(0,weight=1)		
	main_frame=tk.Frame(tabs,background='white')	
	main_frame.rowconfigure(0,weight=1)
	main_frame.columnconfigure(0,weight=1)			
	main_frame.grid(row=0,column=0,sticky=tk.NSEW)
	f=tk.Frame(main_frame,background='white')
	opt=options_module.application_preferences()
	opt_def=options_module.application_preferences()
	o=options_item(f, window,opt_def, 'Preferences')
	return o,main_frame
	
	
		

class options_item(ttk.Treeview):
		
	def __init__(self,frame,window,options,heading):
		self.win=window
		self.options=options
		self.frame_heading=heading
		self.options=options
		self.main_frame=frame
		self.canvas=tk.Canvas(self.main_frame,background='white')
		self.opt_frame=tk.Frame(self.main_frame,background='white')
		self.add_heading()
		ttk.Treeview.__init__(self,self.canvas)
		self.level__dicts=[dict(),dict(),dict()]
		
		self.yscrollbar = ttk.Scrollbar(self.canvas, orient="vertical", command=self.yview)
		self.configure(yscrollcommand=self.yscrollbar.set)
		
		self.xscrollbar = ttk.Scrollbar(self.canvas, orient="horizontal", command=self.xview)
		self.configure(xscrollcommand=self.xscrollbar.set)
		self.tree_construction()
		self.gridding()
		
		self.binding()
		
	def add_heading(self):
		self.name_frame=tk.Frame(self.main_frame,background='white')
		name_lbl=tk.Label(self.name_frame,text=self.frame_heading,background='white')
		name_lbl.grid()
			
		
	def get_script(self):
		scripts=[]
		search_patterns=[]
		d=self.options.__dict__
		for i in d:
			if hasattr(d[i],'value'):
				v=d[i].value
				if v!=self.default_options.__dict__[i].value:
					dtype=d[i].dtype
					if not type(dtype)==list:
						dtype=[dtype]
					if str in dtype:
						if ('\n' in v):
							v=f'"""{v}"""'
						elif '\'' in v:
							v=f'"{v}"'
						else:
							v=f"'{v}'"
					scripts.append(f'options.{i}.set({v})')
					search_patterns.append(fr'options.{i}.set\(([\s\S]*?)\)')
		return scripts,search_patterns
				
		
	def binding(self):
		self.bind('<Double-Button-1>',self.tree_double_click)	
		self.bind('<<TreeviewSelect>>',self.tree_click)	
		self.bind('<Key>',self.key_down)	
		self.bind('<KeyRelease>',self.key_up)		
		
	def tree_construction(self):
		self["columns"]=("one","two")
		self.column("#0", stretch=tk.YES)
		self.column("one", width=50,stretch=tk.YES)
		self.column("two", width=50,stretch=tk.YES)
		self.heading("#0",text="Option",anchor=tk.W)
		self.heading("one", text="value",anchor=tk.W)
		self.heading("two", text="type",anchor=tk.W)	
		self.alt_time=time.perf_counter()
		for k in tags:
			tag_configure(self,k,tags[k])	
		self.tree=dict()	
		self.add_options_to_tree()
		a=0
		
	def gridding(self):
		self.rowconfigure(0,weight=1)
		self.columnconfigure(0,weight=1)
		xscrollbar,yscrollbar=self.xscrollbar,self.yscrollbar
		
		self.main_frame.rowconfigure(0)
		self.main_frame.rowconfigure(1,weight=7)
		self.main_frame.rowconfigure(2,weight=5)
		self.main_frame.columnconfigure(0,weight=1)
		self.canvas.rowconfigure(0,weight=1)
		self.canvas.columnconfigure(0,weight=1)		
		self.opt_frame.rowconfigure(0,weight=1)
		self.opt_frame.columnconfigure(0,weight=1)				
		
		self.main_frame.grid(row=0,column=0,sticky=tk.NSEW)
		self.name_frame.grid(row=0,column=0,sticky=tk.W)
		self.opt_frame.grid(row=2,column=0,sticky='nw')			
		self.canvas.grid(row=1,column=0,sticky=tk.NSEW)	
		xscrollbar.grid(row=1,column=0,sticky='ew')
		yscrollbar.grid(row=0,column=1,sticky='ns')			
		self.grid(row=0,column=0,sticky=tk.NSEW)
		
		
	def key_down(self,event):
		if event.keysym=='Alt_L' or  event.keysym=='Alt_R':
			self.configure(cursor='target')
			self.alt_time=time.perf_counter()
			
			
	def key_up(self,event):
		if event.keysym=='Alt_L' or  event.keysym=='Alt_R':
			self.configure(cursor='arrow')

		
	def tree_double_click(self,event):
		item = self.selection()[0]
		item=self.item(item)['text']
		self.win.main_tabs.insert_current_editor(item)
		
	def tree_click(self,event):
		item = self.selection()
		if len(item)==0:
			return
		item=item[0]
		levels=item.split(';')
		if len(levels)==3:
			parent_itm=';'.join(levels[:-1])
			fname,j,k=levels
			value,vtype=self.item(parent_itm)['values']
			self.item(parent_itm,values=(k,vtype))
			self.item(parent_itm,open=False)
			self.tree[fname][j].option.set(k)
		elif levels[1]!='':#not top level:
			i,j=levels
			if self.item(item)['open']:
				self.item(item,open=False)
			else:
				self.item(item,open=True)
			self.hide_all_frames()
			self.tree[i][j].grid(row=1,column=0)				
			
	def close_all(self):
		for i in self.tree:
			for j in self.tree[i]:
				self.item(j,open=False)	
		
	def add_options_to_tree(self):
		for i in self.options.categories_srtd:
			self.insert('', 1,f"{i};", text=i)
			self.add_node(i,self.options.categories[i])
			self.item(f"{i};",open=True)
		a=0
		
	def hide_all_frames(self):
		for i in self.tree:
			for j in self.tree[i]:
				self.tree[i][j].grid_remove()

	def add_node(self,cat,options):
		d=dict()
		self.tree[cat]=d
		for j in options:
			value=displayvalue(j.value)
			self.insert(f"{cat};",2, f"{cat};{j.name}", text=j.name,values=(value,j.dtype_str))	
			d[j.name]=option_frame(self.opt_frame, j,self,f"{cat};{j.name}")
			self.add_options(j, cat)		

	def add_options(self,option,cat):
		if not option.selection_var:
			return
		for i in range(len(option.permissible_values)):
			desc=''
			if not option.value_description is None:
				desc=option.value_description[i]			
			val= option.permissible_values[i]
			self.insert(f"{cat};{option.name}",i, f"{cat};{option.name};{i}",values=(val,),text=desc, tags=('option',))	
			
	def register_validation(self):
		for i in self.tree:
			for j in self.tree[i]:
				self.tree[i][j].register_validation()
			

def tag_configure(tree,name,d,value=None):
	
	tree.tag_configure(name, foreground=d['fg'])
	tree.tag_configure(name, background=d['bg'])
	tree.tag_configure(name, font=d['font'])	
	if not value is None:
		tree.item(name,value=value)
		
class option_frame(tk.Frame):
	def __init__(self, master, option,option_tree,node_name):
		tk.Frame.__init__(self,master,background='white')
		self.entries=dict()
		self.option_tree=option_tree
		self.node_name=node_name
		self.option=option
		self.lines=dict()
		desc=option.description
		self.desc=tk.Label(self,text=desc,anchor='nw',justify=tk.LEFT,background='white')
		self.desc.grid(row=0,column=0,sticky='nw')		
		if option.is_inputlist:#
			self.cntrl=tk.Frame(self,background='white')
			for i in range(len(option.descr_for_input_boxes)):
				self.add_control_multi(option,self.cntrl,i)
			self.cntrl.grid(row=1,column=0,sticky='nw')
		elif not option.selection_var:
			self.add_control_single(option)
			self.cntrl.grid(row=1,column=0,sticky='nw')
		
	def register_validation(self):
		for i in self.entries:
			self.entries[i].register_validation()
		
			
			
	def add_control_single(self,option):		
		if option.dtype==str:
			self.cntrl=gui_scrolltext.ScrollText(self)
			if not option.value is None:
				self.cntrl.insert('1.0',option.value)
			self.scrolltext_updater=update_options_scrolltext(option, self.cntrl, self.option_tree)
			self.cntrl.bind_to_key_release(self.scrolltext_updater.update)
		else:
			self.cntrl=managed_text(self,option.dtype,option,self.option_tree, self.node_name)
			self.cntrl.text.set(option.value)
			self.entries[self.node_name]=self.cntrl
		

	def add_control_multi(self,option,master,i):		
		line=tk.Frame(self.cntrl,background='white')
		name=self.node_name+str(i)
		line.columnconfigure(0,weight=1)
		line.columnconfigure(1,weight=1)
		desc=option.descr_for_input_boxes[i]
		lbl=tk.Label(line,text=desc,anchor='nw',background='white')
		self.entries[name]=managed_text(line,option.dtype,option,self.option_tree,self.node_name,i)
		if option.value is None:
			self.entries[name].text.set('None')
		else:
			self.entries[name].text.set(str(option.value[i]))
		self.entries[name].grid(row=0,column=2,sticky='nw')
		lbl.grid(row=0,column=0,sticky='nw')
		line.grid(row=i,sticky='nw')
			
			
class update_options_scrolltext:
	def __init__(self,option,scrolltext,option_tree):
		self.option=option
		self.scrolltext=scrolltext
		self.option_tree=option_tree
		
	def update(self):
		P=self.scrolltext.get_all()
		ok=self.option.set(P)
		if not ok:
			return False
		return True
		
		
class managed_text(tk.Entry):
	def __init__(self, master,dtype,option,option_tree,node_name,i=None):
		self.text=tk.StringVar(master)
		tk.Entry.__init__(self,master,textvariable=self.text,validate="key")
		self.option_tree=option_tree
		self.node_name=node_name
		self.dtype=dtype
		self.option=option
		self.i=i
		self.master=master
		
	def register_validation(self):
		vcmd = (self.master.register(self.onValidate),'%d','%P')	
		self.configure(validatecommand=vcmd)
		self.bind('<Return>', self.onEnterKey)
		
	def onEnterKey(self,event):
		self.option_tree.opt_frame.focus()

			
	def onValidate(self,d,P):
		try:
			if P=='':
				return True
			elif P=='None':
				P=None
			dtype=self.option.dtype
			if not(type(dtype)==list or type(dtype)==tuple):
				dtype=[dtype]
			if int in dtype:
				P=int(P)
			elif float in dtype:
				P=float(P)
		except:
			return False
		return self.update_options(P)
		
	def update_options(self,P):
		ok=self.option.set(P,self.i)
		if not ok:
			return
		value=displayvalue(self.option.value)
		self.option_tree.item(self.node_name,values=(value,self.option.dtype_str))
		return ok
	
def displayvalue(value):
	if type(value)==str:
		if '\n' in value:
			return ''	
	return value


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
class options_item:
	def __init__(self,value,description,dtype,name,permissible_values=None,value_description=None, descr_for_input_boxes=[],category='General'):
		"""permissible values can be a vector or a string with an inequality, 
		where %s represents the number, for example "1>%s>0"\n
		if permissible_values is a vector, value_description is a corresponding vector with 
		description of each value in permissible_values"""
		#if permissible_values
		self.description=description
		self.value=value
		self.dtype=dtype
		if type(dtype)==str:
			self.dtype_str=dtype
		elif type(dtype)==list or type(dtype)==tuple:
			self.dtype_str=str(dtype).replace('<class ','').replace('[','').replace(']','').replace('>','').replace("'",'')
		else:
			self.dtype_str=dtype.__name__
			
		self.permissible_values=permissible_values
		self.value_description=value_description
		self.descr_for_input_boxes=descr_for_input_boxes
		self.category=category
		self.name=name
		self.selection_var= len(descr_for_input_boxes)==0 and type(permissible_values)==list
		self.is_inputlist=len(self.descr_for_input_boxes)>0
	
		
		
	def set(self,value,i=None):
		try:
			if not self.valid(value,i):
				return False
		except Exception as e:
			a=self.valid(value,i)
			print(e)
			return False
		if i is None:
			if self.value!=value:
				self.value=value
		else:
			if self.value[i]!=value:
				self.value[i]=value
			else:
				return False
		return True
	
	def valid(self,value,i=None):
		if self.permissible_values is None:
			return (type(value)==self.dtype) or (type(value) in self.dtype)		
		if i is None:
			return self.valid_test(value, self.permissible_values)
		else:
			return self.valid_test(value, self.permissible_values[i])
			
	def valid_test(self,value,permissible):
		if type(permissible)==list or type(permissible)==tuple:
			try:
				if not type(value)==list or type(value)==tuple:
					value=self.dtype(value)
					return value in permissible
				else:
					valid=True
					for i in range(len(value)):
						value[i]=self.dtype(value[i])
						valid=valid*eval(permissible[i] %(value[i],))
			except:
				return False
			return valid
		elif type(permissible)==str:
			return eval(permissible %(value,))
		else:
			print('No method to handle this permissible')
		

		
class options():
	def __init__(self):
		pass
		
		
	def make_category_tree(self):
		opt=self.__dict__
		d=dict()
		keys=np.array(list(opt.keys()))
		keys=keys[keys.argsort()]
		for i in opt:
			if opt[i].category in d:
				d[opt[i].category].append(opt[i])
			else:
				d[opt[i].category]=[opt[i]]
			opt[i].code_name=i
		self.categories=d	
		keys=np.array(list(d.keys()))
		self.categories_srtd=keys[keys.argsort()]



def regression_options():
	#Add option here for it to apear in the "options"-tab. The options are bound
	#to the data sets loaded. Hence, a change in the options here only has effect
	#ON DATA SETS LOADED AFTER THE CHANGE
	self=options()
	
	self.add_intercept				= options_item(True,			"If True, adds intercept if not all ready in the data",
																	bool,'Add intercept', [True,False],['Add intercept','Do not add intercept'],category='Regression')

	self.arguments					= options_item("", 				"A string or dict with a dictionary in python syntax containing the initial arguments." 
																	"An example can be obtained by printing ll.args.args_d"
																	, [str,dict], 'Initial arguments')	

	#self.description				= options_item(None, 			"A description of the project." , 'entry','Description')	
	
	self.do_not_constrain			= options_item(None, 			"The name of a variable of interest \nthat shall not be constrained due to \nmulticollinearity",
													 				[str,type(None)],"Avoid constraint",
																	descr_for_input_boxes=['Variable not to constraint:'])	
	
	self.fixed_random_group_eff		= options_item(2,				'Fixed, random or no group effects', int, 'Group fixed random effect',[0,1,2], 
																	['No effects','Fixed effects','Random effects'],category='Fixed-random effects')
	self.fixed_random_time_eff		= options_item(2,				'Fixed, random or no time effects', int, 'Time fixed random effect',[0,1,2], 
																	['No effects','Fixed effects','Random effects'],category='Fixed-random effects')
	self.fixed_random_variance_eff	= options_item(2,				'Fixed, random or no group effects for variance', int, 'Variance fixed random effects',[0,1,2], 
																	['No effects','Fixed effects','Random effects'],category='Fixed-random effects')
	
	
	self.fixed_random_pre_ARIMA	    = options_item(False,			'Fixed or random effects calculated before ARIMA', bool, 'Calculate fixed/random effects before ARIMA',
																	[True,False],['Before ARIMA','After ARIMA'],category='Fixed-random effects')	
	
	self.fixed_random_in_GARCH      = options_item(False,			'GARCH process is calculated after FE/RE', bool, 'If True, GARCH is calcualted after FE/RE, an not ',
																	[True,False],['GARCH calcualted after FE/RE','GARCH calcualted before FE/RE'],category='Fixed-random effects')	
	
	self.h_function					= options_item(					"def h(e,z):\n"
																	"	e2			=	e**2+1e-5\n"
																	"	h_val		=	np.log(e2)\n"	
																	"	h_e_val		=	2*e/e2\n"
																	"	h_2e_val	=	2/e2-4*e**2/e2**2\n"
																	"	return h_val,h_e_val,h_2e_val,None,None,None\n",	
													
																	"You can supply your own heteroskedasticity function. It must be a function of\n"
																	"residuals e and a shift parameter z that is determined by the maximization procedure\n"
																	"the function must return the value and its derivatives in the following order:\n"
																	"h, dh/de, (d^2)h/de^2, dh/dz, (d^2)h/dz^2,(d^2)h/(dz*de)"
																	, str,"GARCH function",category='Regression')

	self.loadARIMA_GARCH			= options_item(False, 			"Determines whether the ARIMA_GARCH arguments from the previous run should be kept", 
																	bool, 'Load ARIMA_GARCH', [False,True],
																	['No loading',
																	'Load arguments'
																	])	
		
	self.loadargs					= options_item(1, 				"Determines whether the regression arguments from the previous run should be kept", 
																	int, 'Load arguments', [0,1,2],
																	['No loading',
																	'Load from last run of identical model',
																	'Load from last run'
																	])
	
	self.min_group_df				= options_item(1, 				"The smallest permissible number of observations in each group. Must be at least 1", int, 'Minimum degrees of freedom', "%s>0",category='Regression')

	self.minimum_iterations			= options_item(0, 				'Minimum number of iterations in maximization:',
													  				int,"Minimum iterations", "%s>-1")		
	
	#self.pedantic					= options_item(0, 				"Determines how pedantic the maximization shuould be\n"
	#																"If slightly pedantic, the linesearch is allways run twice after the direction is calcualted\n"
	#																"If pedantic, the maximization also tries to fix supected variables\n"
	#																"If very pedantic the maximization also uses brute force as a last resort",  
	#																int,'Pedantic',[0,1,2,3],['Not pedantic', 'Slightly pedantic','Pedantic','Very pedantic'])
	
	self.pool						= options_item(False, 			"True if sample is to be pooled, otherwise False." 
																	"For running a pooled regression",  
																	bool,'Pooling',[True,False],['Pooled','Not Pooled'])
	
	self.pqdkm						= options_item([1,1,0,1,1], 
																	"ARIMA-GARCH parameters:",int, 'ARIMA-GARCH orders',
																	["%s>=0","%s>=0","%s>=0","%s>=0","%s>=0"],
																	descr_for_input_boxes=["Auto Regression order (ARIMA, p)",
																	"Moving Average order (ARIMA, q)",
																	"difference order (ARIMA, d)",
																	"Variance Moving Average order (GARCH, k)",
																	"Variance Auto Regression order (GARCH, m)"],category='Regression')

	self.robustcov_lags_statistics	= options_item([100,30],		"Numer of lags used in calculation of the robust \ncovariance matrix for the time dimension", 
																	 int, 'Robust covariance lags (time)', ["%s>1","%s>1"], 
													 	 	 	 	 descr_for_input_boxes=["# lags in final statistics calulation",
																	 "# lags iterations (smaller saves time)"],
																	 category='Output')

	self.silent						= options_item(False, 			"True if silent mode, otherwise False." 
																	"For running the procedure in a script, where output should be suppressed",  
																	bool,'Silent mode',[True,False],['Silent','Not Silent'])

	self.subtract_means				= options_item(False,			"If True, subtracts the mean of all variables. This may be a remedy for multicollinearity if the mean is not of interest.",
																	bool,'Subtract means', [True,False],['Subtracts the means','Do not subtract the means'],category='Regression')
	
	self.tobit_limits				= options_item([None,None],		"Determines the limits in a tobit regression. "
																	"Element 0 is lower limit and element1 is upper limit. "
																	"If None, the limit is not active", 
																	[float,type(None)], 'Tobit-model limits', 
																	descr_for_input_boxes=['lower limit','upper limit'])

	self.tolerance					= options_item(0.000001, 		"Tolerance. When the maximum absolute value of the gradient divided by the hessian diagonal"
																	"is smaller than the tolerance, the procedure is "
																	"determined to have converged.",
																	float,"Tolerance", "%s>0")	
	
	self.variance_RE_norm			= options_item(0.000005, 		"This parameter determines at which point the log function involved in the variance RE/FE calculations, "
																	"will be extrapolate by a linear function for smaller values",
																	float,"Variance RE/FE normalization point in log function", "%s>0")		
	self.user_constraints			= options_item(None,			"You can add constraints as a dict or as a string in python dictonary syntax.\n",
																	[str,dict], 'User constraints')
	
	self.make_category_tree()
	
	return self


def application_preferences():
	opt=options()
	
	opt.save_datasets	= options_item(True, "If True, all loaded datasets are saved on exit and will reappear when the application is restarted", 
									bool,"Save datasets on exit", [False,True],
									['Save on exit',
									'No thanks'])
	
	opt.n_round	= options_item(4, "Sets the number of digits the results are rounded to", 
									str,"Rounding digits", ['no rounding','0 digits','1 digits','2 digits','3 digits',
																						 '4 digits','5 digits','6 digits','7 digits','8 digits',
																						 '9 digits','10 digits'])
	
	opt.n_digits	= options_item(10, "Sets the maximum number of digits (for scientific format) if 'Rounding digits' is not set (-1)", 
									int,"Number of digits", ['0 digits','1 digits','2 digits','3 digits',
																						 '4 digits','5 digits','6 digits','7 digits','8 digits',
																						 '9 digits','10 digits'])	
	opt.round_scientific	= options_item(True, "Determines if small numbers that are displayed in scientific format shall be rounded", 
								   bool,"Round Scientific", [True,False],['Round Scientific','Do not round scientific'])		
	opt.make_category_tree()
	
	return opt





#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This module calculates statistics and saves it to a file

import numpy as np
from scipy import stats as scstats
from tkinter import font as tkfont
import tkinter as tk
import stat_functions as stat
import functions as fu
import model_parser
STANDARD_LENGTH=8
		


class output:
	def __init__(self,ll,direction,main_msg):
		self.ll=ll
		self.main_msg=main_msg
		self.direction=direction
		self.panel=self.ll.panel
		self.lags=self.panel.options.robustcov_lags_statistics.value[1]
		self.n_variables=self.panel.args.n_args
		self.incr=0
		self.d={'names':np.array(self.panel.args.names_v),
				'count':range(self.n_variables),
				'args':self.ll.args.args_v}
		self.update_after_direction(direction,0)
		self.heading()
		
	def update_after_direction(self,direction,its):
		self.direction=direction		
		self.iterations=its
		self.constraints_printout()
		self.t_stats()
		self.heading()
		
	def update_after_linesearch(self,direction,ll,incr):
		self.direction=direction
		self.ll=ll
		self.incr=incr
		self.d['args']=ll.args.args_v
		self.heading()
		
	def statistics(self):
		return statistics(self.ll)

	def reg_table(self):
		return reg_table_obj(self)
	
	def t_stats(self):
		d=self.d
		direction=self.direction
		panel=self.panel
		
		T=len(d['names'])
		if direction.H is None:
			return
		d['se_robust'],d['se_st']=sandwich(direction,self.lags)
		d['se_robust_oposite'],d['se_st_oposite']=sandwich(direction,self.lags,oposite=True)
		if not (d['se_st_oposite'] is None):
			d['se_robust'][np.isnan(d['se_robust'])]=d['se_robust_oposite'][np.isnan(d['se_robust'])]
			d['se_st'][np.isnan(d['se_st'])]=d['se_st_oposite'][np.isnan(d['se_st'])]
		#d['se_robust_fullsize'],d['se_st_fullsize']=sandwich(direction,self.lags,resize=False)
		no_nan=np.isnan(d['se_robust'])==False
		valid=no_nan
		valid[no_nan]=(d['se_robust'][no_nan]>0)
		d['tstat']=np.array(T*[np.nan])
		d['tsign']=np.array(T*[np.nan])
		d['tstat'][valid]=d['args'][valid]/d['se_robust'][valid]
		d['tsign'][valid]=(1-scstats.t.cdf(np.abs(d['tstat'][valid]),panel.df))#Two sided tests
		d['sign_codes']=get_sign_codes(d['tsign'])
	
	def heading(self):
		CI=self.direction.CI
		n_CI=len(self.direction.mc_problems)
		s=("LL:\t"+str(self.ll.LL)+'  ').ljust(23)
		if not self.incr is None:
			s+=("\tIncrement:  "+ str(self.incr)).ljust(17)+"  "
		else:
			s+=str(" ").ljust(19)
		if not self.iterations is None:
			s+=f"\tIteration:  {str(self.iterations).ljust(7)}"
		if hasattr(self.direction,'HG_ratio'):
			s+=f"\tSingularity problems:  {str(self.direction.singularity_problems).ljust(7)}"
		instr=''
		if not self.panel.input.Z_names is None:
			instr=', '.join(self.panel.input.Z_names[1:])
			instr+="\t"+self.main_msg
		s+=f"\nDependent: {self.panel.input.Y_names[0]}"
		n,T,k=self.panel.X.shape
		s+=f"\tPanel: {self.panel.NT_before_loss} observations,{n} groups and {T} dates"
		if len(instr):
			s+=f"\t\tInstruments: {instr}"		
		s+=f"\nMax condition index: {np.round(self.direction.CI)}\t ({n_CI} caseses where high CI was associated with more than one variable)\n"
		self.heading_str=s		
		
	def constraints_printout(self):
		panel=self.panel
		direction=self.direction
		constr=direction.constr
		weak_mc_dict=direction.weak_mc_dict
		d=self.d
		if not direction.dx_norm is None:
			d['dx_norm']=direction.dx_norm
		T=len(d['names'])
		d['set_to'],d['assco'],d['cause'],d['multicoll']=['']*T,['']*T,['']*T,['']*T
		if constr is None:
			return
		c=constr.fixed
		for i in c:
			d['set_to'][i]=c[i].value_str
			d['assco'][i]=c[i].assco_name
			d['cause'][i]=c[i].cause	
			
		c=constr.intervals
		for i in c:
			if not c[i].intervalbound is None:
				d['set_to'][i]=c[i].intervalbound
				d['assco'][i]='NA'
				d['cause'][i]=c[i].cause		
				
		for i in weak_mc_dict:#adding associates of non-severe multicollinearity
			d['multicoll'][i]='|'
			d['assco'][i]=panel.args.names_v[weak_mc_dict[i][0]]	

	
class reg_table_obj(dict):
	def __init__(self,output):
		dict.__init__(self)
		self.d=output.d
		self.Y_names=output.ll.panel.input.Y_names
		self.args=output.ll.args.dict_string
		self.n_variables=output.n_variables
		self.heading=output.heading_str
		self.footer=f"\n\nSignificance codes: '=0.1, *=0.05, **=0.01, ***=0.001,    |=collinear\n\n{output.ll.err_msg}"	
	
	def table(self,n_digits,brackets,fmt,stacked, show_direction, show_constraints):
		include_cols,llength=self.get_cols(stacked, show_direction, show_constraints)
		if fmt=='INTERNAL':
			self.X=None
			return str(self.args),None
		self.include_cols=include_cols
		self.n_cols=len(include_cols)		
		for a, l,is_string,name,neg,just,sep,default_digits in pr:		
			self[a]=column(self.d,a,l, is_string, name, neg, just, sep, default_digits,self.n_variables)
		self.X=self.output_matrix(n_digits,brackets)
		s=format_table(self.X, include_cols,fmt,f'Regression on {self.Y_names[0]}',self.heading,self.footer)
		return s,llength

	
	def output_matrix(self,digits,brackets):
		structured=False
		for i in range(self.n_cols):
			if type(self.include_cols[i])==list:
				structured=True
				break
		if structured:
			return self.output_matrix_structured(digits, brackets)
		else:
			return self.output_matrix_flat(digits, brackets)
		
		
	def output_matrix_structured(self,digits,brackets):
		X=[['']*self.n_cols for i in range(3*(self.n_variables+1)-1)]
		for i in range(self.n_cols):
			a=self.include_cols[i]
			if type(a)==list:
				h=self[a[0]].name.replace(':',' ')
				if brackets=='[':
					X[0][i]=f"{h}[{self[a[1]].name}]:"
				elif brackets=='(':
					X[0][i]=f"{h}({self[a[1]].name}):"
				else:
					X[0][i]=f"{h}/{self[a[1]].name}:"
				v=[self[a[j]].values(digits) for j in range(3)]
				for j in range(self.n_variables):
					X[(j+1)*3-1][i]=v[0][j]
					if brackets=='[':
						X[(j+1)*3][i]=f"[{v[1][j]}]{v[2][j]}"
					elif brackets=='(':
						X[(j+1)*3][i]=f"({v[1][j]}){v[2][j]}"
					else:
						X[(j+1)*3][i]=f"{v[1][j]}{v[2][j]}"
			else:
				X[0][i]=self[a].name
				v=self[a].values(digits)
				for j in range(self.n_variables):
					X[(j+1)*3-1][i]=v[j]
		return X	
	
	def output_matrix_flat(self,digits,brackets):
		X=[['']*self.n_cols for i in range(self.n_variables+1)]
		for i in range(self.n_cols):
			a=self.include_cols[i]
			X[0][i]=self[a].name
			v=self[a].values(digits)
			for j in range(self.n_variables):
				X[j+1][i]=v[j]
		return X	
	

	def get_cols(self,stacked,
				show_direction,
				show_constraints):
		"prints a single regression"
		dx_col=[]
		llength=9
		if show_direction:
			dx_col=['dx_norm']
		else:
			llength-=1
		mcoll_col=[]
		if show_constraints:
			mcoll_col=[ 'multicoll','assco','set_to', 'cause']
		else:
			llength-=2		
		if stacked:
			cols=['count','names', ['args','se_robust', 'sign_codes']] + dx_col + ['tstat', 'tsign'] + mcoll_col
		else:
			cols=['count','names', 'args','se_robust', 'sign_codes'] + dx_col + ['tstat', 'tsign'] + mcoll_col		
		return cols,llength
	
	
class column:
	def __init__(self,d,a,l,is_string,name,neg,just,sep,default_digits,n_variables):		
		self.length=l
		self.is_string=is_string
		self.name=name
		self.default_digits=default_digits
		self.neg_allowed=neg
		self.justification=just
		self.tab_sep=sep
		self.n_variables=n_variables
		if a in d:
			self.exists=True
			self.input=d[a]
		else:
			self.exists=False
			self.input=[' - ']*self.n_variables		
		
	def values(self,digits):
		try:
			if self.length is None:
				if digits=='SCI':
					return self.input
				else:
					return np.round(self.input,digits)
			return np.round(self.input,self.length)
		except:
			if self.length is None:
				return self.input
			else:
				return np.array([str(i).ljust(self.length)[:self.length] for i in self.input])

			
def get_preferences(output_gui):
	try:
		pref=output_gui.window.right_tabs.preferences.options
		return pref
	except:
		return
	
def sandwich(direction,lags,oposite=False,resize=True):
	panel=direction.panel
	H,G,idx=reduce_size(direction,oposite,resize)
	lags=lags+panel.lost_obs
	try:
		hessin=np.linalg.inv(-H)
	except Exception as e:
		print(e)
		return None,None
	se_robust,se,V=stat.robust_se(panel,lags,hessin,G)
	se_robust,se,V=expand_x(se_robust, idx),expand_x(se, idx),expand_x(V, idx,True)
	return se_robust,se

def reduce_size(direction,oposite,resize):
	H=direction.H
	G=direction.G
	if (G is None) or (H is None):
		return
	m=len(H)
	if not resize:
		return H,G,np.ones(m,dtype=bool)
	weak_mc_dict=direction.weak_mc_dict.keys()
	constr=list(direction.constr.fixed.keys())	
	if oposite:
		weak_mc_dict=[direction.weak_mc_dict[i][0] for i in direction.weak_mc_dict]
		constr=[]
		for i in direction.constr.fixed:
			if not direction.constr.fixed[i].assco_ix is None:
				constr.append(direction.constr.fixed[i].assco_ix)
	for i in weak_mc_dict:
		if not i in constr:
			constr.append(i)
	idx=np.ones(m,dtype=bool)
	if len(constr)>0:#removing fixed constraints from the matrix
		idx[constr]=False
		H=H[idx][:,idx]
		G=G[:,:,idx]
	return H,G,idx

def expand_x(x,idx,matrix=False):
	m=len(idx)
	if matrix:
		x_full=np.zeros((m,m))
		x_full[:]=np.nan
		ref=np.arange(m)[idx]
		for i in range(len(x)):
			try:
				x_full[ref[i],idx]=x[i]
				x_full[idx,ref[i]]=x[i]
			except:
				a=0
	else:
		x_full=np.zeros(m)
		x_full[:]=np.nan
		x_full[idx]=x
	return x_full
	
	
def get_sign_codes(tsign):
	sc=[]
	for i in tsign:
		if np.isnan(i):
			sc.append(i)
		elif i<0.001:
			sc.append('***')
		elif i<0.01:
			sc.append('** ')
		elif i<0.05:
			sc.append('*  ')
		elif i<0.1:
			sc.append("'  ")
		else:
			sc.append('')
	sc=np.array(sc,dtype='<U3')
	return sc

def remove_illegal_signs(name):
	illegals=['#', 	'<', 	'$', 	'+', 
	          '%', 	'>', 	'!', 	'`', 
	          '&', 	'*', 	'‘', 	'|', 
	          '{', 	'?', 	'“', 	'=', 
	          '}', 	'/', 	':', 	
	          '\\', 	'b']
	for i in illegals:
		if i in name:
			name=name.replace(i,'_')
	return name
		
	
class statistics:
	def __init__(self,ll):
		ll.standardize()
		panel=ll.panel
		self.df=panel.df
		self.N,self.T,self.k=panel.X.shape
		self.Rsq_st, self.Rsqadj_st, self.LL_ratio,self.LL_ratio_OLS=stat.goodness_of_fit(ll,True)	
		self.Rsq, self.Rsqadj, self.LL_ratio,self.LL_ratio_OLS=stat.goodness_of_fit(ll,False)	
		self.no_ac_prob,self.rhos,self.RSqAC=stat.breusch_godfrey_test(panel,ll,10)
		self.DW=stat.DurbinWatson(panel,ll)
		self.norm_prob=stat.JB_normality_test(ll.e_norm,panel)
		self.ADF_stat,self.c1,self.c5=stat.adf_test(panel,ll,10)
		self.df_str=self.gen_df_str(panel)	
		self.instruments=panel.input.Z_names[1:]
		self.pqdkm=panel.pqdkm
		
	def gen_df_str(self,panel):
		summary=f"""
  SAMPLE SIZE SUMMARY:
\tOriginal sample size\t\t:\t{orig_size}
\tSample size after filtering\t\t:\t{panel.NT_before_loss}
\tDegrees of freedom\t\t:\t{panel.df}
\tNumber of IDs\t\t:\t{self.N:,}
\tNumber of dates (maximum)\t\t:\t{self.T}\n"""		
			
		group_rmv=f"""
  REMOVED GROUPS BECAUSE OF TOO FEW OBSERVATIONS:
\tObservations per group lost because of
\tA)\tARIMA/GARCH\t:\t{panel.lost_obs}
\tB)\tMin # of obs in user preferences:\t:\t{panel.options.min_group_df.value}
\tMin # observations required (A+B)\t\t:\t{panel.lost_obs+panel.options.min_group_df.value}\n
\tGroups removed
\tA)\tTotal # of groups\t:\t{len(panel.idincl)}
\tB)\t# of groups removed\t:\t{sum(panel.idincl==False)}
\t# of groups remaining (A-B)\t\t:\t{sum(panel.idincl==True)}

\t# of observations removed\t\t:\t{panel.input.X.shape[0]-panel.NT_before_loss}\n"""
		
		
		s=f"""
{summary}
{group_rmv}
  DEGREES OF FREEDOM:
\tA)\tSample size\t:\t{panel.NT_before_loss}
\tB)\tObservations lost to
\t\tGARCH/ARIMA\t:\t{panel.tot_lost_obs}	
\tRandom/Fixed Effects in
\tC)\tMean process\t:\t{panel.number_of_RE_coef}
\tD)\tVariance process\t:\t{panel.number_of_RE_coef_in_variance}
\tE)\tNumber of coefficients in
\t\tRegression\t:\t{panel.args.n_args:,}
\tDegrees of freedom (A-B-C-D-E)\t\t:\t{panel.df}\n\n"""
		return s
	
	def gen_mod_fit(self,n_digits):
		return f"""	
\tLL-ratio\t\t:\t{round(self.LL_ratio,n_digits)}
\tR-squared (from observed data)\t\t:\t{round(self.Rsq*100,2)}%
\tAdjusted R-squared (from observed data)\t\t:\t{round(self.Rsqadj*100,2)}%
\tR-squared (from normalized data)\t\t:\t{round(self.Rsq_st*100,2)}%
\tAdjusted R-squared (from normalized data)\t\t:\t{round(self.Rsqadj_st*100,2)}%

\t("Normalized data" means that the data is adjusted with the estimated ARIMA-GARCH parameters and random/fixed effects.)

\tDurbin-Watson statistic:\t\t:\t{round(self.DW,2)}
\tBreusch-Godfrey test\t\t:\t{round(self.no_ac_prob*100,n_digits)}% \t(significance, probability of no auto correlation)
\tJarque–Bera test for normality\t\t:\t{round(self.norm_prob*100,n_digits)}% \t(significance, probability of normality)\n
"""		


	def adf_str(self,n_digits):
			
		if not self.ADF_stat=='NA':
			if self.ADF_stat<self.c1:
				self.ADF_res="Unit root rejected at 1%"
			elif self.ADF_stat<self.c5:
				self.ADF_res="Unit root rejected at 5%"
			else:
				self.ADF_res="Unit root not rejected"		
			adf=f"""
\tAugmented Dicky-Fuller (ADF)        
\t\tTest statistic\t:\t{round(self.ADF_stat,n_digits)}
\t\t1% critical value\t:\t{round(self.c1,n_digits)}
\t\t5% critical value\t:\t{round(self.c5,n_digits)}
\t\tResult\t:\t\t{self.ADF_res}
			"""
		else:
			adf="Unable to calculate ADF"
		if self.df<1:
			s+="""
\tWARNING: All your degrees of freedom (df) has been consumed, so statistics cannot be computed.
\tyou can increase df by for example turning off random/fixed effects """
		return adf
	
	
def get_tab_stops(X,f):
	f = tkfont.Font(font=f)
	m_len = f.measure("m")
	counter=2*m_len
	tabs=[f"{counter}",tk.NUMERIC]
	r,c=np.array(X).shape
	for i in range(c):
		t=1
		num_max=0
		for j in range(r):
			s=str(X[j][i])
			if '.' in s:
				a=s.split('.')
				num_max=max((len(a[0]),num_max))
			t=max((f.measure(X[j][i])+(num_max+2)*m_len,t))
		counter+=t
		tabs.extend([f"{counter}",tk.NUMERIC])			
	return tabs

l=STANDARD_LENGTH
#python variable name,	length,		is string,  display name,		neg. values,	justification	next tab space		round digits (None=no rounding,-1=set by user)
pr=[
		['count',		2,			False,		'',					False,			'right', 		2,					None],
		['names',		None,		True,		'Variable names:',	False,			'right', 		2, 					None],
		['args',		None,		False,		'Coef:',				True,			'right', 		2, 					-1],
		['se_robust',	None,		False,		'rob.SE',		True,			'right', 		3, 					-1],
		['sign_codes',	5,			True,		'',					False,			'left', 		2, 					-1],
		['dx_norm',		None,		False,		'direction:',		True,			'right', 		2, 					None],
		['tstat',		2,			False,		't-stat.:',			True,			'right', 		2, 					2],
		['tsign',		None,		False,		'p-value:',			False,			'right', 		2, 					3],
		['multicoll',	1,			True,		'',					False,			'left', 		2, 					None],
		['assco',		20,			True,		'collinear with',	False,			'center', 		2, 					None],
		['set_to',		6,			True,		'set to',			False,			'center', 		2, 					None],
		['cause',		50,			True,		'cause',			False,			'right', 		2, 					None]]		

		
def format_table(X,cols,fmt,heading,head,tail):
	if fmt=='NORMAL':
		return head+format_normal(X,[1],cols)+tail
	if fmt=='LATEX':
		return head+format_latex(X,cols,heading)+tail
	if fmt=='HTML':
		return format_html(X,cols,heading,head)+tail	
	
	
def format_normal(X,add_rows=[],cols=[]):
	p=''
	if 'multicoll' in cols:
		constr_pos=cols.index('multicoll')+1
		p="\t"*constr_pos+"constraints:".center(38)
	p+="\n"
	for i in range(len(X)):
		p+='\n'*(i in add_rows)
		p+='\n'
		for j in range(len(X[0])):
			p+=f'\t{X[i][j]}'
		
	return p	


def format_latex(X,cols,heading):
	X=np.array(X,dtype='U128')
	n,k=X.shape
	p="""
\\begin{table}[ht]
\\caption{%s} 
\\centering
\\begin{tabular}{""" %(heading,)
	p+=' c'*k+' }\n\\hline'
	p+='\t'+' &\t'.join(X[0])+'\\\\\n\\hline\\hline'
	for i in range(1,len(X)):
		p+='\t'+ ' &\t'.join(X[i])+'\\\\\n'
	p+="""
\hline %inserts single line
\end{tabular}
\label{table:nonlin} % is used to refer this table in the text
\end{table}"""
	return p	

def format_html(X,cols,heading,head):
	X=np.array(X,dtype='U128')
	n,k=X.shape
	head=head.replace('\n','<br>')
	head=head.replace('\t','&nbsp;'*4)
	p=f"""
	<h1>{heading}</h1>
	<p>{head}</p>
	<p><table>"""
	p+='\t</tr><th>'+'\t</th><th>'.join(X[0])+'</th></tr>\n'
	for i in range(1,len(X)):
		p+='\t</tr><td>'+'\t</td><td>'.join(X[i])+'</td></tr>\n'
	p+='</table></p>'
	return p		
	
alphabet='abcdefghijklmnopqrstuvwxyz'
class join_table(dict):
	"""Creates a  joint table of several regressions with columns of the join_table_column class.
	See join_table_column for data handling."""
	def __init__(self,args,varnames=[]):
		dict.__init__(self)
		self.names_category_list=list([list(i) for i in args.names_category_list])#making a copy
		k=0
		for i in varnames:
			if i in self.names_category_list[0]:
				k=self.names_category_list[0].index(i)+1
			else:
				self.names_category_list[0].insert(k,i)
				k+=1
		self.names_v=[itm for s in self.names_category_list for itm in s]#flattening
		
	def update(self,ll,stats,desc):
		if not desc in self:
			for i in range(len(ll.args.names_category_list)):
				for j in ll.args.names_category_list[i]:
					if not j in self.names_category_list[i]:
						self.names_category_list[i].append(j)
			self.names_v=[itm for s in self.names_category_list for itm in s]#flattening
		self[desc]=join_table_column(stats, ll)
		
		
	def make_table(self, stacked, brackets,digits,caption):
		keys=list(self.keys())
		k=len(keys)
		n=len(self.names_v)
		if stacked:
			X=[['' for j in range(2+k)] for i in range(4+3*n)]
			for i in range(n):
				X[3*i+1][1]=self.names_v[i]		
				X[3*i+1][0]=i
			X[1+3*n][1]='Log likelihood'
			X[2+3*n][1]='Degrees of freedom'	
			X[3+3*n][1]='Adjusted R-squared'	
		else:
			X=[['' for j in range(2+2*k)] for i in range(4+n)]
			for i in range(n):
				X[i+1][1]=self.names_v[i]	
				X[i+1][0]=i
			X[1+n][1]='Log likelihood'
			X[2+n][1]='Degrees of freedom'		
			X[3+n][1]='Adjusted R-squared'	
		for i in range(k):
			self.make_column(i,keys[i],X,stacked, brackets,digits,caption)
		s=format_normal(X,[1,(1+stacked*2)*n+1,(1+stacked*2)*n+4])
		s+=f"\n\nSignificance codes: '=0.1, *=0.05, **=0.01, ***=0.001,    |=collinear\n"
		max_mod=0
		models=[]
		for i in range(len(keys)):
			key=self[keys[i]]
			p,q,d,k,m=key.pqdkm
			models.append(f"\n{alphabet[i]}: {keys[i]}")
			max_mod=max(len(models[i]),max_mod)
		for i in range(len(keys)):
			s+=models[i].ljust(max_mod+2)
			if len(key.instruments):
				s+=f"\tInstruments: {key.instruments}"
			s+=f"\tARIMA({p},{d},{q})-GARCH({k},{m})"
		return s,X
			
	def make_column(self,col,key,X,stacked, brackets,digits,caption):
		if not 'se_robust' in self[key].stats:
			return
		
		if caption=='JOINED LONG':
			X[0][(2-stacked)*col+2]+=f"{self[key].Y_name} ({alphabet[col]})"
		else:
			X[0][(2-stacked)*col+2]=alphabet[col]
		n=len(self.names_v)
		m=len(self[key].args.names_v)
		ix=[self.names_v.index(i) for i in self[key].args.names_v]
		se=np.round(self[key].stats['se_robust'],digits)
		sgn=self[key].stats['sign_codes']
		args=np.round(self[key].args.args_v,digits)
		if brackets=='[':
			se_sgn=[f"[{se[i]}]{sgn[i]}" for i in range(m)]
		elif brackets=='(':
			se_sgn=[f"({se[i]}){sgn[i]}" for i in range(m)]
		else:
			se_sgn=[f"{se[i]}{sgn[i]}" for i in range(m)]				
		if stacked:
			for i in range(m):
				X[3*ix[i]+1][col+2]=args[i]
				X[3*ix[i]+2][col+2]=se_sgn[i]
			X[1+3*n][col+2]=fu.round_sign_digits(self[key].LL,5,1)
			X[2+3*n][col+2]=self[key].df	
			X[3+3*n][col+2]=f"{round(self[key].Rsqadj*100,1)}%"
		else:
			for i in range(m):
				X[ix[i]+1][col*2+2]=args[i]
				X[ix[i]+1][col*2+3]=se_sgn[i]		
			X[1+n][col*2+3]=fu.round_sign_digits(self[key].LL,5,1)
			X[2+n][col*2+3]=self[key].df
			X[3+n][col*2+3]=f"{round(self[key].Rsqadj*100,1)}%"
	


class join_table_column:
	def __init__(self,stats,ll):
		panel=ll.panel
		self.stats=stats
		self.LL=ll.LL
		self.df=ll.panel.df
		self.args=ll.args
		self.Rsq, self.Rsqadj, self.LL_ratio,self.LL_ratio_OLS=stat.goodness_of_fit(ll,True)
		self.instruments=panel.input.Z_names[1:]
		self.pqdkm=panel.pqdkm		
		self.Y_name=panel.input.Y_names
		


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import subprocess
try:
	import win32clipboard
except:
	mswin=False

def copy_wondows():
	CF_RTF = win32clipboard.RegisterClipboardFormat("Rich Text Format")
	rtf = bytearray("{\\rtf1\\ansi\\deff0 {\\pard This is a test\\par}}", 'utf8')
	win32clipboard.OpenClipboard(0)
	win32clipboard.EmptyClipboard()
	win32clipboard.SetClipboardData(CF_RTF, rtf)
	win32clipboard.CloseClipboard()
	
def copy_linux(string):
	#could this work for windows too?
	if str(type(string)) == "<class 'str'>":
		string = bytearray(string, 'utf8')
	subprocess.Popen(['xclip', '-selection', 'clipboard', '-t', 'text/rtf'], stdin=subprocess.PIPE).communicate(string)
	
text = r"{\rtf1\ansi\deff0 {\b This} is some text\row}"
copy_rtf(text)	


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import tkinter as tk

class CreateToolTip(object):
	"""
	create a tooltip for a given widget
	"""
	def __init__(self, widget, text='widget info'):
		self.wraplength = 180   #pixels
		self.widget = widget
		self.text = text
		self.widget.bind("<Enter>", self.enter)
		self.widget.bind("<Leave>", self.leave)
		self.widget.bind("<ButtonPress>", self.leave)
		self.id = None
		self.tw = None

	def enter(self, event=None):
		self.showtip(event)

	def leave(self, event=None):
		self.hidetip()

	def showtip(self, event=None):
		x = y = 0
		x, y, cx, cy = self.widget.bbox("insert")
		x += self.widget.winfo_rootx() + 25
		y += self.widget.winfo_rooty() + 25
		# creates a toplevel window
		self.tw = tk.Toplevel(self.widget)
		# Leaves only the label and removes the app window
		self.tw.wm_overrideredirect(True)
		self.tw.wm_geometry("+%d+%d" % (x, y))
		label = tk.Label(self.tw, text=self.text, justify='left',
		                 background="#ffffff", relief='solid', borderwidth=1,
		               wraplength = self.wraplength)
		label.pack(ipadx=1)

	def hidetip(self):
		tw = self.tw
		self.tw= None
		if tw:
			tw.destroy()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import tkinter as tk
from tkinter import ttk
from multiprocessing import pool
import sys
from gui import gui_charts
from gui import gui_functions as guif
import functions as fu
import os
from gui import gui_buttons
import functions as fu
from gui import gui_right_tabs
from gui import gui_scrolltext
from gui import gui_main_tabs
import tempstore
import numpy as np
import traceback
FONT_SIZE=10	
FONT_WIDTH=FONT_SIZE*0.35	
LINE_HEIGHT=1.54

GRAPH_IMG_WIDTH=0.35
GRAPH_IMG_HEIGHT=0.85



class window(tk.Tk):
	def __init__(self):
		tk.Tk.__init__(self)
		self.title("Paneltime")
		self.define_styles()
		self.mc=None
		self.geometry('%sx%s+%s+%s' %(self.winfo_screenwidth(),self.winfo_screenheight()-75,-5,0))
		if 'nt'==os.name:
			iconpath=os.path.join(fu.currentdir(),'paneltime.ico')
			self.iconbitmap(iconpath)	
			self.iconpath=iconpath
		else:
			self.iconpath=None
		#self.add_menu()
		self.add_panes()
		#self.main_pane=self
		self.main_pane.rowconfigure(0, weight=1)
		self.main_pane.columnconfigure(0, weight=80,uniform="fred")
		self.main_pane.columnconfigure(1)
		self.main_pane.columnconfigure(2, weight=20,uniform="fred")
		self.add_main_frames()
		self.add_delimiters()
		self.output = gui_scrolltext.ScrollText(self.frm_left,format_text=False)
		self.output.grid(row=2, column=0,sticky=tk.NSEW)
		self.data=datastore(self)
		self.autosave()
		self.right_tabs=gui_right_tabs.right_tab_widget(self)
		self.main_tabs=gui_main_tabs.main_tabs(self)
		self.locals=dict()
		self.globals={'window':self,'data':self.right_tabs.data_tree.datasets}
		sys.stdout=stdout_redir(self.output)
		sys.stderr=stdout_redir(self.output)
		self.protocol("WM_DELETE_WINDOW", self.on_closing)
		
	def autosave(self):
		if fu.get_idle_duration()>600:
			self.data.save()
		self.after(600000,self.autosave)
	
	def exec(self,source):
		try:
			exec(source,self.globals,self.locals)
			return True
		except Exception as e:
			traceback.print_exc()
		return False
	
	def define_styles(self):
		style = ttk.Style()
		style.configure("TFrame", background='white')
		style.configure("TNotebook",background='white')

	def add_panes(self):
		self.rowconfigure(0)
		self.rowconfigure(1,weight=1,uniform="fred")
		self.rowconfigure(2)
		self.columnconfigure(0,weight=1)
		
		self.button_pane=tk.Frame(self,height=10,background="white")
		#self.buttons=gui_buttons.buttons(self)
		self.main_pane=tk.Canvas(self,background='pink')
		self.bottom_bar=tk.Frame(self,background='white',height=25)
		self.about=tk.Label(self.bottom_bar,text='This sofware is a scientific work and should be cited if used for scientific purposes. Cite Sirnes (2020)',background='white')	
		self.about.grid(row=0,column=0,sticky=tk.W)
		self.button_pane.grid(row=0,column=0,sticky=tk.EW)	
		self.main_pane.grid(row=1,column=0,sticky=tk.NSEW)	
		self.bottom_bar.grid(row=2,column=0,sticky=tk.EW)
		
	def do_nothing(self,event):
		pass
		
	def add_delimiters(self):
		self.delimiter_v=tk.Frame(self.frm_left,background="dark grey",height=5,cursor='sb_v_double_arrow')
		self.delimiter_v.grid(row=1,column=0,sticky=tk.EW)
		self.delimiter_v.bind("<ButtonRelease-1>", self.vertical_resize)
		
		self.delimiter_h=tk.Frame(self.main_pane,background="dark grey",width=5,cursor='sb_h_double_arrow')
		self.delimiter_h.grid(row=0,column=1,sticky=tk.NS)
		self.delimiter_h.bind("<ButtonRelease-1>", self.horizontal_resize)	
		
	def add_main_frames(self):
		left_weight=80
		self.frm_left = tk.Frame(self.main_pane,background='green')
		
		self.frm_left.rowconfigure(0, weight=left_weight,uniform="fred")
		self.frm_left.rowconfigure(1)
		self.frm_left.rowconfigure(2, weight=100-left_weight,uniform="fred")
		self.frm_left.columnconfigure(0, weight=1)
		self.frm_left.grid(row=0,column=0,sticky=tk.NSEW)
		
		self.frm_right = tk.Frame(self.main_pane,background='white')
		self.frm_right.rowconfigure(0)
		self.frm_right.rowconfigure(1,weight=1)
		self.frm_right.columnconfigure(0,weight=1)
		self.frm_right.grid(row=0,column=2,sticky=tk.NSEW)	
		
		self.main_frames_weight=left_weight
		
		#self.box_border_frm=tk.Frame(self.frm_left,background="red",height=100)
	
	def vertical_resize(self,event):
		self.pack_propagate(0)
		new_y=self.output.winfo_y()+event.y
		y=int(100*(new_y)/(self.winfo_height()))
		self.frm_left.rowconfigure(0, weight=y,uniform="fred")
		self.frm_left.rowconfigure(2, weight=100-y,uniform="fred")	
		
	def horizontal_resize(self,event):
		self.pack_propagate(0)
		new_x=self.frm_right.winfo_x()+event.x
		x=int(100*new_x/(self.winfo_width()))
		self.main_pane.columnconfigure(0, weight=max((x,0)))
		self.main_pane.columnconfigure(2, weight=max((100-x,0)))	
		
	def add_menu(self):
		menubar = tk.Menu(self)
		filemenu = tk.Menu(menubar, tearoff=0)
		menubar.add_cascade(label="File", menu=filemenu)

		openmenu = tk.Menu(menubar, tearoff=0)
		filemenu.add_cascade(label="Open data", menu=openmenu)
		openmenu.add_command(label="Data text file", command=self.donothing)
		openmenu.add_command(label="Sql connection", command=self.donothing)
		filemenu.add_command(label="Open project", command=self.donothing)
		filemenu.add_separator()
		filemenu.add_command(label="Save project", command=self.donothing)
		filemenu.add_command(label="Save project as", command=self.donothing)
		filemenu.add_separator()
		filemenu.add_command(label="Settings", command=self.donothing)
		filemenu.add_separator()
		filemenu.add_command(label="Quit", command=self.donothing)		
		
		self.settingsmenu = tk.Menu(menubar, tearoff=0)
		menubar.add_cascade(label="Regression", menu=self.settingsmenu)
		self.settingsmenu.add_command(label="Run Ctrl+R", command=self.donothing)
		self.settingsmenu.add_command(label="Abort Ctrl+A", command=self.abort)
		filemenu.add_separator()
		self.settingsmenu.add_command(label="Settings", command=self.donothing)
		filemenu.add_separator()
		self.settingsmenu.add_command(label="Scatter plots raw", command=self.show_scatter,state="disabled")
		self.settingsmenu.add_command(label="Scatter plots normalized", command=self.show_scatter_norm,state="disabled")
		
		self.config(menu=menubar)
			
	def donothing(self):
		pass

		
	def abort(self):
		if self.btn_abort.cget('relief')==tk.RAISED:
			self.btn_abort.config(relief=tk.SUNKEN)
		else:
			self.btn_abort.config(relief=tk.RAISED)
		
	def done(self,x):
		sys.stdout=sys.__stdout__
		self.pool.terminate()		
		self.destroy()
		self.quit()
		
	def on_closing(self):
		if self.right_tabs.preferences.options.save_datasets.value:
			tempstore.save_obj(tempstore.fname_datasets,self.right_tabs.data_tree.datasets)		
		self.data.save()
		exit()			
	
	def get(self):
		return self.process.get()

		
class stdout_redir():
	def __init__(self,textbox):
		self.textbox = textbox

	def write(self,string):
		self.textbox.insert('end', string)
		self.textbox.see('end')
		

class datastore(dict):
	def __init__(self,window):
		d=tempstore.load_obj(tempstore.fname_window)
		if d is None or type(d)!=dict:
			dict.__init__(self)
		else:
			dict.__init__(self,d)
		self.dict_default=dict()
		self.dict_default['sql_str']="SELECT * FROM TABLE <table>"
		self.dict_default['conn_str']=def_conn_str
		self.win=window
		self['current path']=os.getcwd()
		self['current json path']=os.getcwd()

		
	def get(self,key,default=None):
		try:
			v=self[key]
			if v is None:
				self[key]=default
			return self[key]
		except:
			self[key]=default
		return self[key]
		
	def __getitem__(self,key):
		try:
			return dict.__getitem__(self,key)
		except:
			self[key]=self.dict_default[key]
			return dict.__getitem__(self,key)

		
	def save(self):
		self.win.main_tabs._tabs.save_all_in_temp()
		tempstore.save_obj(tempstore.fname_window,dict(self))
		tempstore.save_zip()

				
	
def_conn_str="""conn = pymysql.connect(host='<hostname>', \n
\t\t\tuser='<username>', 
\t\t\tpasswd='<password>', 
\t\t\tdb='<dbname>')	"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
import os
import multi_core
import traceback
import datetime
import pickle
import gc

def main(t,initcommand,s_id,fpath):
	fname=os.path.join(fpath,'slaves/%s.txt' %(s_id,))
	f=open(fname,'w',1)
	d=dict()
	exec(initcommand,globals(),d)
	d_list=list(d.keys())
	holdbacks=[]
	while 1:
		(msg,obj) = t.receive()
		response=None
		if msg=='kill':
			sys.exit()
			response=True
		elif msg=='dictionary':#a dictionary to be used in the batch will be passed
			d_new=obj
			add_to_dict(d,d_new)
			d_list=list(d.keys())
			response=True
		elif msg=='filetransfer':
			ftr=open(obj, "rb")
			u= pickle.Unpickler(ftr)
			d_new=u.load()		
			add_to_dict(d,d_new)
			d_list=list(d.keys())
			response=True
			ftr.close()
		elif msg=='remote expression valuation':
			expr,ret_var=obj
			sys.stdout = f
			exec(expr,globals(),d)
			sys.stdout = sys.__stdout__	
			response={ret_var:d[ret_var]}
		elif msg=='remote recieve':
			response=d[obj]
		elif msg=='expression evaluation':	
			sys.stdout = f
			exec(obj,globals(),d)
			sys.stdout = sys.__stdout__
			response=release_dict(d,d_list,holdbacks)	
		elif msg=='holdbacks':
			holdbacks.extend(obj)		
		t.send(response)		
		if  msg=='expression evaluation':#remove response dict after sending
			for i in response:
				d.pop(i)
		gc.collect()
		

def add_to_dict(to_dict,from_dict):
	for i in from_dict:
		to_dict[i]=from_dict[i]

def write(f,txt):
	f.write(str(txt)+'\n')
	f.flush()

def release_dict(d,d_list,holdbacks):
	"""Ensures that only new variables are returned"""
	response=dict()
	for i in d:
		if (not i in d_list) and (not i in holdbacks):
			response[i]=d[i]	
	return response

try: 

	t=multi_core.transact(sys.stdin, sys.stdout)
	#Handshake:
	t.send(os.getpid())
	msg,(initcommand,s_id,fpath)=t.receive()
	fname=os.path.join(fpath,'slave_errors.txt')
	f=open(fname,'w')
	#Wait for instructions:
	main(t,initcommand,s_id,fpath)
except Exception as e:
	f.write('SID: %s      TIME:%s \n' %(s_id,datetime.datetime.now()))
	traceback.print_exc(file=f)
	f.flush()
	f.close()
	raise RuntimeError(e)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import tkinter as tk
from tkinter import ttk
from gui import gui_charts
import functions as fu
import time
from gui import gui_scrolltext
from gui import gui_script_handling
from gui import gui_tooltip
from gui import gui_sql
import options as options_module
from tkinter import filedialog
import json
import os
import tempstore
import time
import paneltime as pt
import numpy as np
import pandas as pd

font='Arial 9 '
tags=dict()
tags['dependent']={'fg':'#025714','bg':'#e6eaf0','font':font+'bold','short':'Y'}
tags['independent']={'fg':'#053480','bg':'#e6eaf0','font':font+'bold','short':'X'}
tags['time variable']={'fg':'#690580','bg':'#e6eaf0','font':font+'bold','short':'T'}
tags['id variable']={'fg':'#910101','bg':'#e6eaf0','font':font+'bold','short':'ID'}
tags['het.sc._factors']={'fg':'#029ea3','bg':'#e6eaf0','font':font+'bold','short':'HF'}
selected={'bg':'#ebfbfc'}
unselected={'fg':'black','bg':'white','font':font,'short':''}
#for correct sorting:
tags_list=['dependent','independent','time variable','id variable','het.sc._factors']

	
class data_objects(ttk.Treeview):
	def __init__(self,tabs,window,right_tabs):
		s = ttk.Style()
		s.configure('new.TFrame', background='white',font=font)	
		
		self.tabs=tabs
		self.clicked=False
		self.right_tabs=right_tabs
		self.win=window
		self.main_frame=tk.Frame(tabs)
		self.add_buttons()
		self.canvas=tk.Canvas(self.main_frame)
		ttk.Treeview.__init__(self,self.canvas,style='new.TFrame')
		self.dbase_img=self._img = tk.PhotoImage(file= fu.join(os.path.dirname(__file__),['img','database.png']),master=self)
		self.var_img=self._img = tk.PhotoImage(file= fu.join(os.path.dirname(__file__),['img','variable.png']),master=self)		
		self.level__dicts=[dict(),dict(),dict()]
		
		yscrollbar = ttk.Scrollbar(self.canvas, orient="vertical", command=self.yview)
		self.configure(yscrollcommand=yscrollbar.set)
		
		xscrollbar = ttk.Scrollbar(self.canvas, orient="horizontal", command=self.xview)
		self.configure(xscrollcommand=xscrollbar.set)
		
		self.gridding(xscrollbar,yscrollbar)
		self.tree_construction()
		self.binding()
			

		self.load()
		self.script=''
		self.expand_time=time.perf_counter()
		self.gui_sql=None
		self.import_buttons=None

		
		
	def add_buttons(self):
		bgcolor='white'
		self.button_frame=tk.Frame(self.main_frame,height=22,background=bgcolor)
		self.button_img=dict()
		
		self.button_add=self.add_button(self.button_frame,'add.png',self.add,'Opens a new dataset',bgcolor)
		self.button_save=self.add_button(self.button_frame,'save_as.png',self.save,'Save current dataset as a csv file',bgcolor)
		self.button_delete=self.add_button(self.button_frame,'delete.png',self.delete_dataset,'Remove current dataset',bgcolor)
		self.button_edit_vars=self.add_button(self.button_frame,'edit_variables.png',self.edit_variables,'Edit variables in current data set',bgcolor)
		self.button_import_script=self.add_button(self.button_frame,'import_script.png',self.insert_load_script,'Display import script',bgcolor)
		
	def insert_load_script(self):
		dataset=self.get_selected_dataset()
		dataset.get_script_editor(self.win.main_tabs)

		
	def add_button(self,master,img,command,tooltip,bgcolor,size=22):
		self.button_img[img]= tk.PhotoImage(file =  fu.join(os.path.dirname(__file__),['img',img]),master=self.button_frame)
		btn=tk.Button(master, image = self.button_img[img],command=command, 
								   highlightthickness=0,bd=0,height=size,width=size,compound='left',background=bgcolor)
		gui_tooltip.CreateToolTip(btn,tooltip)
		return btn	

	def add(self):
		if self.import_buttons is None:
			self.add_import_buttons()
			return
		if self.import_buttons.state()=='normal':
			self.import_buttons.withdraw()
		else:
			self.import_buttons.deiconify()
		
	def add_import_buttons(self):
		bgcolor='white'
		x = y = 0
		x, y, cx, cy = self.button_add.bbox("insert")
		x += self.button_add.winfo_rootx() + 25
		y += self.button_add.winfo_rooty() + 25
		# creates a toplevel window
		self.import_buttons = tk.Toplevel(self.button_add)
		# Leaves only the label and removes the app window
		self.import_buttons.wm_overrideredirect(True)
		self.import_buttons.geometry('+%d+%d' % (x, y))
		frm = tk.Frame(self.import_buttons, background="#ffffff", relief='flat')
		self.button_add_sql=self.add_button(frm,'Table_sql.png',self.open_sql,'Add dataset from an SQL database',bgcolor,40)
		self.button_add_file=self.add_button(frm,'Table_File.png',self.open_file,'Open a data file',bgcolor, 40)
		self.button_add_json=self.add_button(frm,'Table_JSON.png',self.open_json,'Open a JSON file',bgcolor,40)
		self.button_add_copy=self.add_button(frm,'Table_copy.png',self.copy,'Copy current data set for running system regression',bgcolor,40)
		

		self.button_add_sql.grid(row=0,column=0,padx=10,pady=10)
		self.button_add_file.grid(row=1,column=0,padx=10,pady=10)
		self.button_add_json.grid(row=0,column=1,padx=10,pady=10)
		self.button_add_copy.grid(row=1,column=1,padx=10,pady=10)
		frm.grid()
		

	def open_json(self):
		p=self.win.data['current json path']
		filename = filedialog.askopenfilename(initialdir=p,title="Open data file",
			filetypes = (("JSON", "*.json"),) )
		if not filename: 
			return
		p,f=os.path.split(filename)
		self.win.data['current json path']=p		
		exe_str=f"from paneltime import *\nload_json('{filename}')"
		
		exe_str=f"""from paneltime import *\n
data=dict()\ndata['{f}']=load_json('{filename}')"""
		res=self.win.exec(exe_str)
		if res==False:
			return		
		df=self.win.locals['data'][f]
		tree=self.right_tabs.data_tree
		tree.datasets.add(tree,f,df,filename,exe_str,self.win.main_tabs)
	
	def copy(self):
		pass	
	
	def open_file(self):
		p=self.win.data['current path']
		filename = filedialog.askopenfilename(initialdir=p,title="Open data file",
			filetypes = (("CSV", "*.csv")
	        ,("text", "*.txt")
			,("All files", "*.*") ))
		if not filename: 
			return
		p,f=os.path.split(filename)
		self.win.data['current path']=p
		exe_str=f"""from paneltime import *\n
data=dict()\ndata['{f}']=load('{filename}')"""
		res=self.win.exec(exe_str)
		if res==False:
			return
		df=self.win.locals['data'][f]
		tree=self.right_tabs.data_tree
		tree.datasets.add(tree,f,df,filename,exe_str)
		
	def open_sql(self):
		if self.gui_sql is None:
			self.gui_sql=gui_sql.sql_query(self.win,self)
		else:
			self.gui_sql.show()
			
	def edit_variables(self):
		dataset=self.get_selected_dataset()
		dataset.get_data_editor(self.main_tabs)
		
	def load(self):
		self.datasets=tempstore.load_obj(tempstore.fname_datasets)
		if self.datasets is None:
			self.datasets=datasets()
			return
		self.datasets.make_trees(self)
		s=self.win.data.get('selected data')
		try:
			self.focus_set()
			self.focus(s)
		except:
			self.win.data['selected data']=';'
			return
			
	def save(self):
		p=self.win.data['current path']
		filename = filedialog.asksaveasfilename(initialdir=p,title="Save",
			filetypes = (("CSV", "*.csv"),), defaultextension=True)
		if not filename: 
			return
		self.path=filename
		p,f=os.path.split(filename)
		self.win.data['current path']=p
		dataset=self.get_selected_dataset()
		X=[]
		for i in dataset:
			if type(dataset[i])==np.ndarray:
				if len(dataset[i].shape)==2:
					X.append(np.concatenate(([[i]],dataset[i])))
		if len(X)==0:
			print('No data in dataset')
			return
		X=np.concatenate(X,1)
		np.savetxt(filename, X,fmt='%s',delimiter=',')
		print('dataset saved')
	
	
	def delete_dataset(self):
		df=self.get_selected_dataset()
		if df is None:
			return
		self.datasets.delete(df.name,self)
		
	def binding(self):
		self.bind('<Double-Button-1>',self.tree_double_click)	
		self.bind('<<TreeviewSelect>>',self.tree_select)	
		self.bind('<Button-1>',self.tree_click)	
		self.bind('<Key>',self.key_down)	
		self.bind('<KeyRelease>',self.key_up)		
		
	def tree_construction(self):
		self["columns"]=("one","two")
		self.column("#0", stretch=tk.YES)
		self.column("one", width=15,stretch=tk.YES)
		self.column("two", width=75,stretch=tk.YES)
		self.heading("#0",text="Name",anchor=tk.W)
		self.heading("one", text="",anchor=tk.W)
		self.heading("two", text="type",anchor=tk.W)	
		self.alt_time=time.perf_counter()
		for k in tags_list:
			self.tag_configure_node(k,tags[k],False)		
		
	def gridding(self,xscrollbar,yscrollbar):
		self.main_frame.rowconfigure(0)
		self.main_frame.rowconfigure(1,weight=1,uniform="fred")	
		self.main_frame.columnconfigure(0,weight=1)
		
		self.rowconfigure(0,weight=1)
		self.columnconfigure(0,weight=1)
		self.tabs.rowconfigure(0,weight=1)
		self.tabs.columnconfigure(0,weight=1)
		
		
		self.button_frame.grid(row=0,sticky=tk.EW)	
		
		pad=8
		self.button_add.grid(row=0,column=0,padx=pad,pady=pad)
		self.button_delete.grid(row=0,column=1,padx=pad,pady=pad)
		self.button_edit_vars.grid(row=0,column=2,padx=pad,pady=pad)
		self.button_save.grid(row=0,column=3,padx=pad,pady=pad)
		self.button_import_script.grid(row=0,column=4,padx=pad,pady=pad)
		
		
		self.canvas.rowconfigure(0,weight=1)
		self.canvas.columnconfigure(0,weight=1)		
		self.main_frame.grid(row=0,column=0,sticky=tk.NSEW)
		self.canvas.grid(row=1,column=0,sticky=tk.NSEW)	
		
		xscrollbar.grid(row=1,column=0,sticky='ew')
		yscrollbar.grid(row=0,column=1,sticky='ns')	
		self.grid(row=0,column=0,sticky=tk.NSEW)	

		
	def key_down(self,event):
		if event.keysym=='Alt_L' or  event.keysym=='Alt_R':
			self.configure(cursor='target')
			self.alt_time=time.perf_counter()
			
			
	def key_up(self,event):
		if event.keysym=='Alt_L' or  event.keysym=='Alt_R':
			self.configure(cursor='arrow')

		
	def tree_double_click(self,event):
		item = self.selection()[0]
		levels=item.split(';')
		if levels[1]=='' or len(levels)==3:
			return
		ds=self.datasets[levels[0]]
		txt=self.item(item)['text']
		editor=ds.get_data_editor(self.win.main_tabs)
		editor.widget.write(txt)
		editor.frame.focus()
		
	def tree_click(self,event=None):
		item=self.selected_item()
		self.clicked=True
		
	def tree_select(self,event=None):
		
		item=self.selected_item()
		dname,vname,sel=(item.split(';')+[''])[:3]
		if (dname,vname,sel)==('', '', ''):
			return
		self.win.data['selected data']=dname+';'
		if vname!='':#not top level
			if sel!='':
				self.var_defined(dname,vname,sel)
			else:
				self.var_clicked(item)
			gui_script_handling.edit_exe_script(self.win,self.datasets[dname])
		else:
			self.expand_collaps_parent(item)
		self.select_selected(dname+';')
		self.clicked=False
		self.right_tabs.sel_data.set(dname)
		
	def expand_collaps_parent(self,item):
		if not self.clicked:
			return
		if not item.split(';')[1]=='':
			return
		if self.item(item)['open']:
			self.item(item,open=False)
		else:
			self.item(item,open=True)	
		

	def select_selected(self,parent):
		for i in self.get_children():
			self.tag_configure(i, background=unselected['bg'])
			for j in self.get_children(i):
				t=self.tag_configure(j)
				if t['background']==selected['bg']:
					self.tag_configure(j, background=unselected['bg'])
		self.tag_configure(parent, background=selected['bg'])
		for i in self.get_children(parent):
			t=self.tag_configure(i)
			if t['background']==unselected['bg'] or  t['background']=='':	
				self.tag_configure(i, background=selected['bg'])

					
	def var_clicked(self,item):
		item_obj=self.item(item)
		short,vtype=item_obj['values']
		t=self.tag_configure(item)	
		if item_obj['open']:
			if t['font']!=unselected['font']:
				self.tag_configure_node(item,unselected)
			else:
				self.close_all()
		else:
			if time.perf_counter()-self.alt_time<0.1:#alt pressed
				if short=='':
					self.tag_configure_node(item,tags['independent'])
				else:
					self.tag_configure_node(item,unselected)
			else:
				self.close_all()
				self.item(item,open=True)
				
	def var_defined(self,dname,vname,sel):
		parent_itm=dname+';'+vname
		short,vtype=self.item(parent_itm)['values']
		s=tags[sel]['short']
		if s=='Y' or s=='T' or s=='ID':#only one of these are allowed
			for i in self.datasets.desc[dname].nodes:
				short_i,vtype_i=self.item(i)['values']
				if s==short_i:
					self.tag_configure_node(i,unselected)
		if short==s:#Click on same
			self.tag_configure_node(parent_itm,unselected)
		else:
			self.tag_configure_node(parent_itm,tags[sel])
		self.item(parent_itm,open=False)
		
				
	def close_all(self):
		for i in self.datasets:
			for j in self.datasets.desc[i].nodes:
				self.item(j,open=False)
				
	def selected_item(self):
		item = self.selection()
		if len(item)==0:
			return
		if len(item)!=1:
			item=[self.win.data.get('selected data',item[0])]
		item=item[0]	
		return item
		
	def get_selected_dataset(self):
		item=self.selected_item()
		if item is None:
			return
		itm=item.split(';')[0]
		return self.datasets[itm]	
	
	def tag_configure_node(self,name,d,addvalue=True):
		self.tag_configure(name, foreground=d['fg'])
		self.tag_configure(name, background=d['bg'])
		self.tag_configure(name, font=d['font'])	
		if not addvalue:
			return
		short,dtype=self.item(name)['values']
		self.item(name,values=(d['short'],dtype))
		fname=name.split(';')[0]
		self.datasets.desc[fname].nodes[name]=d

def np_type(name,df):
	x=df[name]
	return x.dtype

class dataset_description():
	def __init__(self,datasets,name,source,import_script,tree):
		self.source=source
		self.data_import_script=import_script
		self.exe_editor=None
		self.edit_editor=None
		self.script_editor=None
		self.name=name
		self.nodes={}
		self.datasets=datasets
		self.options=options_module.regression_options()
		self.auto_update=True
		
		
	def generate_exe_script(self,tree):
		X=[]
		HF=[]
		d=dict()
		d['Y'],d['ID'],d['T'],d['HF']='','','',''
		for i in self.nodes:
			fname,j=i.split(";")
			if tree.item(i)['values'][0]=='X':
				X.append(j)
			elif tree.item(i)['values'][0]=='HF':
				HF.append(j)			
			else:
				d[tree.item(i)['values'][0]]=f'{j}'

		args=[f"'{d['Y']}~{'+'.join(X)}'\n\t",f"data['{self.name}']"]
		if len(HF):
			d['HF']=f"[{','.join(HF)}]"
		for i in ['ID','T','HF']:
			if d[i]!='':
				args.append(f"{i}='{d[i]}'")
		return f"execute({','.join(args)})"
		
			
class datasets(dict):
	def __init__(self):
		dict.__init__(self)
		self.desc={}
	
	def add(self,tree,name,data_dict,source,import_script):
		self[name]=pd.DataFrame(data_dict)
		self.desc[name]=dataset_description(self,name, source, import_script,tree)
		self.make_trees(tree)
		tree.item(f"{name};",open=True)
		
	def make_trees(self,tree):
		for i in self:
			self.make_tree(i,tree)	
			
	def make_tree(self,name,tree):
		try:
			tree.insert('', 1,f"{name};", text=' '+name,tags=(name+';',),image=tree.dbase_img)
		except tk.TclError:
			tree.delete(f"{name};")
			tree.insert('', 1,f"{name};", text=' '+name,tags=(name+';',),image=tree.dbase_img)
		self.add_nodes(tree,name)
		tree.main_frame.focus()
		tree.selection_add(f"{name};")
		tree.item(name+';',open=True)
		self.expand_check=[True,name+';']
		
	def delete(self,item,tree):
		self.__delitem__(item)
		tree.delete(item+';')
		
				
	def add_nodes(self,tree,name,d=None):
		df=self[name]
		nodes=self.desc[name].nodes
		item_list=list(df)
		if not d is None:
			if len(item_list)!=len(d.keys()):
				for i in d:
					df[i]=d[i]
		item_list.sort(key=lambda v: v.upper())
		n=len(item_list)
		n_tags=len(tags_list)
		for j in item_list:
			nptype=np_type(j,df)
			nodename=f"{name};{j}"
			tree.insert(f"{name};",n, nodename, text=' '+j,values=('',nptype),tags=(nodename,), image=tree.var_img)	
			if not nodename in nodes:
				nodes[nodename]=unselected
			tree.tag_configure_node(nodename,nodes[nodename])
			for k in tags_list:
				tree.insert(nodename,n_tags, f"{name};{j};{k}",values=('',tags[k]['short']), text=k,tags=(k,))
	
		




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import tkinter as tk
import keyword
import numpy as np
import re

font0="Courier 10"
ret_chr=['\r','\n']

class ScrollText(tk.Canvas):
	def __init__(self,master,readonly=False,text=None,format_text=True,name=None, window=None,height=None,x_bar=True):
		tk.Canvas.__init__(self,master,bg='white')
		self.rowconfigure(0,weight=1)
		self.columnconfigure(0,weight=1)		
		
		self.yscrollbar = tk.Scrollbar(self)
		self.yscrollbar.set(0,0)			
		
		self.xscrollbar=None
		if x_bar==True:
			self.xscrollbar = tk.Scrollbar(self,orient='horizontal')
			self.xscrollbar.set(0,0)

		self.master=master
		self.text_box = CustomText(self, wrap = tk.NONE,xscroll = self.xscrollbar,
								   yscroll = self.yscrollbar,undo=True,format_text=format_text,
								   name=name,window=window,height=height)			
		self.text_box.config(tabs='1c')
		if not self.xscrollbar is None:
			self.xscrollbar.config(command = self.text_box.xview)
			self.xscrollbar.grid(row=1,column=0,sticky='ew')
		
		self.yscrollbar.config(command = self.text_box.yview)
		self.yscrollbar.grid(row=0,column=1,sticky='ns')
		
		self.text_box.grid(row=0,column=0,sticky=tk.NSEW,padx=(15,0),pady=(5,0))
		
		self.readonly=readonly
		if not text is None:
			self.replace_all(text)
		if readonly:
			self.text_box.configure(state='disabled')
			
	def bind_to_key_release(self,func):
		self.text_box.after_key_release=func

		
	def get(self,index1,index2):
		return self.text_box.get(index1,index2)
	
	def get_all(self):
		return self.get('1.0',tk.END)
		
	def delete(self,index1,index2):
		if self.readonly:
			self.text_box.configure(state='normal')	
		self.text_box.delete(index1,index2)
		if self.readonly:
			self.text_box.configure(state='disabled')

	def insert(self,index1,chars,index2=None):
		if self.readonly:
			self.text_box.configure(state='normal')
		if not index2 is None:
			self.text_box.delete(index1, index2)
		self.text_box.insert(index1,chars)
		self.text_box.replace
		if self.readonly:
			self.text_box.configure(state='disabled')
		self.text_box.key_released()
		
	def write(self,chars):
		if self.readonly:
			return
		self.insert(tk.INSERT,chars)
		
	def see(self,index):
		self.text_box.see(index)
		
	def replace_all(self,string):
		if string is None:
			string=""
		if self.readonly:
			self.text_box.configure(state='normal')		
		if not self.xscrollbar is None:
			x=self.xscrollbar.get()
		y=self.yscrollbar.get()
		lines=string.split('\n')
		n=int(self.text_box.index('end').split('.')[0])-1
		if len(lines)==n and False:#attemting to reduce flickering, but does not seem to help
			for i in range(n):
				s=f'{i+1}.0',f'{i+1}.end'
				self.text_box.delete(s[0],s[1])
				self.text_box.insert(s[0],lines[i])
		else:
			self.text_box.delete('1.0',tk.END)
			self.text_box.insert('1.0',string)	
		if not self.xscrollbar is None:
			self.xscrollbar.set(*x)
			if x[1]>x[0]:
				self.text_box.xview(tk.MOVETO,x[0]/(x[1]-x[0]))

		self.yscrollbar.set(*y)		
		if y[1]>y[0]:
			self.text_box.yview(tk.MOVETO,y[0]/(y[1]-y[0]))
			
		if self.readonly:
			self.text_box.configure(state='disabled')
			
		self.text_box.key_released()
			
	



class CustomText(tk.Text):

	def __init__(self,master, wrap,undo,xscroll=None,yscroll=None,format_text=True,name=None,window=None,height=None):
		font='Courier'
		size=10
		
		yscrollcommand=None
		xscrollcommand=None
		if yscroll is None: 
			yscrollcommand=yscroll.set
		if not xscroll is None:
			xscrollcommand=xscroll.set
		
		tk.Text.__init__(self, master,wrap=wrap, 
						 xscrollcommand=xscrollcommand,yscrollcommand=yscrollcommand,undo=undo,bd=0,height=height)	
		self.master=master
		self.configure(font=(font,size,'normal'))
		self.bind('<KeyRelease>', self.key_released)
		self.bind('<KeyPress>', self.key_pressed)
		self.tag_configure('quote', foreground='dark red')
		self.tag_configure('builtins', foreground='#51769e')
		self.tag_configure('keyword', foreground='#0a00bf')
		self.tag_configure('comment', foreground='#00a619')
		self.tag_configure('definition', foreground='#008a5a')
		self.tag_configure('bold', font=(font,size,'bold'))
		self.tag_configure('normal', font=(font,size,'normal'))
		self.tag_configure('black', foreground='black')
		self.define_keywords()
		self.format_text=format_text
		self.pressed_key=''
		self.released_ignore_key=False
		self.name=name
		self.win=window
		self.after_key_release=None
		
		
		
	def define_keywords(self):
		kwlist=np.array(keyword.kwlist)
		kwlensrt=np.array([len(i) for i in keyword.kwlist]).argsort()
		self.kwrds=list(kwlist[kwlensrt])
		builtins=list(dir(__builtins__))+builtin_functions
		b=[]
		for i in builtins:
			if not i in self.kwrds:
				b.append(i)
		self.builtins=b
		False 
				
		
	def key_pressed(self,event):
		try:
			self.pressed_key=event.keysym
		except:
			pass
		
	
	def key_released(self,event=None):
		if not self.format_text:return
		if self.pressed_key in ignore_press_keys:
			self.pressed_key=''
			return
		if not event is None:
			if ((self.released_ignore_key==True) 
				and (not event.keysym in ignore_press_keys)):
				self.released_ignore_key=False
				if event.keysym!='v':#format text after pasting
					return
			if event.keysym in ignore_press_keys:
				self.released_ignore_key=True
			if (event.keysym in ignore_keys):
				return
		for tag in self.tag_names():
			self.tag_remove(tag,'1.0','end')	
			
		self.highlight_pattern(r"\"\"\"(.*?)\"\"\"", 'quote')
		self.highlight_pattern(r"\"(.*?)\"", 'quote')	
		self.highlight_pattern(r"'(.*?)'", 'quote')
		self.highlight_pattern_multiline(r"\"\"\"([\s\S]*?)\"\"\"", 'quote')

		for i in self.kwrds:
			self.highlight_pattern(r"\m(%s)\M" %(i,), 'keyword',tag2='bold')		
		for i in self.builtins:
			self.highlight_pattern(r"\m(%s)\M" %(i,), 'builtins')				
		self.highlight_pattern(r"def (.*?)\(", 'definition',addstart=4,subtractend=1,tag2='bold')
		
		self.highlight_pattern(r"#(.*?)\r", 'comment',end='end-1c')
		self.highlight_pattern(r"#(.*?)\n", 'comment',end='end-1c')
		if not self.after_key_release is None:
			self.after_key_release()

		
	def highlight_pattern(self, pattern, tag, start="1.0", end="end",
	                      regexp=True,tag2=None,addstart=0,subtractend=0):


		indicies=self.search_pattern(pattern, start=start, end=end, regexp=regexp, addstart=addstart, subtractend=subtractend)
		if len(indicies)==0:
			return
		for index1,index2 in indicies:
			if not len(self.tag_names(index1)):
				self.tag_add(tag, index1, index2)
				if not tag2 is None:
					self.tag_add(tag2, index1, index2)
				
	def highlight_pattern_multiline(self, pattern, tag, start="1.0", end="end",
	                      regexp=True,tag2=None,addstart=0,subtractend=0):

		s=self.get('1.0',tk.END)
		start=0
		while True:
			m=re.search(pattern,s[start:])
			if m is None:break
			index1=self.pos_to_index(m.start()+start,s)
			index2=self.pos_to_index(m.end()+start,s)
			self.tag_add(tag, index1, index2)
			if not tag2 is None:
				self.tag_add(tag2, index1, index2)	
			start=m.end()+start
			
	def search_pattern(self, pattern, start="1.0", end="end",
	                      regexp=True,addstart=0,subtractend=0):


		start = self.index(start)
		end = self.index(end)
		self.mark_set("matchStart", start)
		self.mark_set("matchEnd", start)
		self.mark_set("searchLimit", end)

		count = tk.IntVar(master=self)
		indicies=[]
		while True:
			index1 = self.search(pattern, "matchEnd","searchLimit",
								count=count, regexp=regexp)
			if index1 == "": break
			n=count.get()-subtractend
			if n <= 0: break # degenerate pattern which matches zero-length strings
			index2="%s+%sc" % (index1, n)
			if addstart>0:
				index1=f"{index1}+{addstart}c"
			self.mark_set("matchStart", index1)
			self.mark_set("matchEnd", index2)			
			indicies.append((index1,index2))
			
			
		return indicies

			
		
	def pos_to_index(self,pos,string):
		s=str(string[:pos+1])
		if pos+1>len(s):
			pos=len(s)-1
		lines=s.split('\n')
		if len(lines)==1:
			return f"1.{len(s)}" 
		chars=lines[-1]
		index=f"{len(lines)}.{len(chars)}"
		return index
	
			

ignore_press_keys=['Up','Down','Right','Left','Prior','Next','Tab','Control_L','Control_R','Alt_R','Alt_L']
ignore_keys=ignore_press_keys+['Shift_L','Shift_R']


builtin_functions=["abs", "all", "any", "ascii", "bin", "bool", "bytearray", "bytes", "callable", "chr", "classmethod", 
				   "compile", "complex", "delattr", "dict", "dir", "divmod", "enumerate", "eval", "exec", "filter", 
				   "float", "format", "frozenset", "getattr", "globals", "hasattr", "hash", "help", "hex", "id", 
				   "input", "int", "isinstance", "issubclass", "iter", "len", "list", "locals", "map", "max", "memoryview", 
				   "min", "next", "object", "oct", "open", "ord", "pow", "print", "property", "range", "repr", "reversed", 
				   "round", "set", "setattr", "slice", "sorted", "staticmethod", "str", "sum", "super", "tuple", "type", 
				   "vars", "zip", "__import__", ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import calculus
import calculus_functions as cf
import numpy as np
import constraints
import loglikelihood as logl
from scipy import stats
import sys


class direction:
	def __init__(self,panel,mp,channel):
		self.progress_bar=channel.set_progress
		self.gradient=calculus.gradient(panel,self.progress_bar)
		self.hessian=calculus.hessian(panel,self.gradient,self.progress_bar)
		self.panel=panel
		self.constr=None
		self.hessian_num=None
		self.do_shocks=True
		self.input_old=None
		self.CI=0
		self.mp=mp
		self.dx_norm=None
		self.dx=None
		self.H=None
		self.G=None
		self.g=None
		self.weak_mc_dict=[]
		self.mc_problems=[]
		self.H_correl_problem=False
		self.singularity_problems=False
		self.record=[]
		self.Hess_correction=1
		self.g_rec=[]
		self.overflow_problem=False


	def calculate(self,ll,its,msg,increment,lmbda,rev):
		if ll.LL is None:
			raise RuntimeError("Error in LL calculation: %s" %(ll.err_msg,))
		self.ll=ll
		self.its=its
		self.lmbda=lmbda
		self.has_reversed_directions=rev
		self.increment=increment
		self.constr_old=self.constr
		self.constr=constraints.constraints(self,ll.args,its)
		self.constr.add_static_constraints(self.overflow_problem)			
		self.calc_gradient()
		self.calc_hessian()
		self.calc_HG_ratio()
		
		self.constr.add_dynamic_constraints(self)	
		self.CI=self.constr.CI
		self.H_correl_problem=self.constr.H_correl_problem	
		self.mc_problems=self.constr.mc_problems
		self.weak_mc_dict=self.constr.weak_mc_dict
		
		self.singularity_problems=(len(self.mc_problems)>0) or self.H_correl_problem
		
	def set(self,mp,args,constraints=None):
		if not constraints is None:
			self.constr.add_custom_constraints(constraints,
										 cause='pedantic constraint',
										 clear=True,args=args)
		mp.send_dict_by_file({'constr':self.constr})
		self.dx=solve(self.constr,self.H, self.g, args.args_v)
		#self.dx=solve_delete(self.constr,self.H, self.g, args.args_v)	
		self.dx_norm=self.normalize(self.dx)
		try:
			self.dx_unconstr=self.normalize(solve(None,self.H, self.g, args.args_v))
		except:
			self.dx_unconstr=self.dx_norm
	
	def include(self,all=False):
		include=np.array(self.panel.args.n_args*[True])
		if all:
			return include
		include[list(self.constr)]=False
		return include	

	def calc_gradient(self):
		
		ll=self.ll
		DLL_e=-(ll.e_RE*ll.v_inv)*self.panel.included[3]
		dLL_lnv=-0.5*(self.panel.included[3]-(ll.e_REsq*ll.v_inv)*self.panel.included[3])	
		dLL_lnv*=ll.dlnv_pos
		self.LL_gradient_tobit(ll, DLL_e, dLL_lnv)
		self.g,self.G=self.gradient.get(ll,DLL_e,dLL_lnv,return_G=True)
		if len(self.g_rec):
			if not np.all(self.g_rec[-1]==self.g):
				self.g_rec.append(self.g)
		else:
			self.g_rec.append(self.g)
			
		max_history=3
		start=2
		max_history=min((max_history+start,len(self.g_rec)+start))
		self.g_w=[np.mean(self.g_rec[-i:],0) for i in range(start,max_history)]
		#self.g_w=[np.mean(self.g_rec[-i:],0) for i in range(start,max_history)]
		
	
	def LL_gradient_tobit(self,ll,DLL_e,dLL_lnv):
		g=[1,-1]
		self.f=[None,None]
		self.f_F=[None,None]
		for i in [0,1]:
			if self.panel.tobit_active[i]:
				I=self.panel.tobit_I[i]
				self.f[i]=stats.norm.pdf(g[i]*ll.e_norm[I])
				self.f_F[i]=(ll.F[i]!=0)*self.f[i]/(ll.F[i]+(ll.F[i]==0))
				self.v_inv05=ll.v_inv**0.5
				DLL_e[I]=g[i]*self.f_F[i]*self.v_inv05[I]
				dLL_lnv[I]=-0.5*DLL_e[I]*ll.e_RE[I]
				a=0
				

	def LL_hessian_tobit(self,ll,d2LL_de2,d2LL_dln_de,d2LL_dln2):
		g=[1,-1]
		if sum(self.panel.tobit_active)==0:
			return
		self.f=[None,None]
		e1s1=ll.e_norm
		e2s2=ll.e_REsq*ll.v_inv
		e3s3=e2s2*e1s1
		e1s2=e1s1*self.v_inv05
		e1s3=e1s1*ll.v_inv
		e2s3=e2s2*self.v_inv05
		f_F=self.f_F
		for i in [0,1]:
			if self.panel.tobit_active[i]:
				I=self.panel.tobit_I[i]
				f_F2=self.f_F[i]**2
				d2LL_de2[I]=      -(g[i]*f_F[i]*e1s3[I] + f_F2*ll.v_inv[I])
				d2LL_dln_de[I] =   0.5*(f_F2*e1s2[I]  +  g[i]*f_F[i]*(e2s3[I]-self.v_inv05[I]))
				d2LL_dln2[I] =     0.25*(f_F2*e2s2[I]  +  g[i]*f_F[i]*(e1s1[I]-e3s3[I]))

	def calc_hessian(self):
		ll=self.ll
		d2LL_de2=-ll.v_inv*self.panel.included[3]
		d2LL_dln_de=ll.e_RE*ll.v_inv*self.panel.included[3]
		d2LL_dln_de*=ll.dlnv_pos
		d2LL_dln2=-0.5*ll.e_REsq*ll.v_inv*self.panel.included[3]	
		d2LL_dln2*=ll.dlnv_pos
		self.LL_hessian_tobit(ll, d2LL_de2, d2LL_dln_de, d2LL_dln2)
		self.H=self.hessian.get(ll,self.mp,d2LL_de2,d2LL_dln_de,d2LL_dln2)
		self.handle_overflow()
		d=np.diag(self.H)
		self.CI=constraints.decomposition(self.H)[0][-1]
		det=np.linalg.det(self.H)
		self.record.append([self.ll.LL,self.increment,self.CI,det]+list(d))

		if (self.overflow_problem 
			or np.any(d==0) 
			#or (np.any(d>=0) and det>0) 
			or (self.CI>1e+18 and self.increment<1)
			or self.increment<1e-3
			):
			self.H=0.5*(self.H+np.diag(d))
			self.progress_bar(text='Hessian diagonal doubled because it had positive or zero elements')

	def handle_overflow(self):
		self.overflow_problem=False
		try:
			self.H_det=np.linalg.det(self.H)
			return
		except (RuntimeWarning,OverflowError) as e:
			self.overflow_problem=True
			self.H=np.diag(np.ones(len(self.H)))
			self.g[np.abs(self.g)>1e+30]=0
			self.H_det=np.linalg.det(self.H)	
			return
		d=np.array(np.diag(self.H))	
		log_d=np.log(np.abs(d))
		if np.sum(log_d)<max_ln_sum:
			self.H=np.diag(d)
			self.H_det=np.linalg.det(self.H)
			return
		d_srtd=np.sort(np.abs(d))[::-1]
		for i in range(len(d_srtd)-1):
			d[np.abs(d)>=d_srtd[i]]=d_srtd[i+1]
			if np.sum(np.log(np.abs(d)))<max_ln_sum:
				self.H=np.diag(d)
				self.H_det=np.linalg.det(self.H)				
				return 

	
	def calc_HG_ratio(self):
		self.H_g_approx=-cf.dot(self.G,self.G)
		self.HG_ratio=-1
		try:
			detGG=np.linalg.det(self.H_g_approx)
		except Exception as e:
			return
		try:
			detH=np.linalg.det(self.H)
		except Exception as e:
			return		
		try:
			if detGG*detH>0:	
				self.HG_ratio=np.log(detGG/detH)
		except:
			return


	
	def init_ll(self,args):
		self.constr=constraints.constraints(self,args)
		self.constr.add_static_constraints(False)			
		try:
			args=args.args_v
		except:
			pass#args must be a vector
		for i in self.constr.fixed:
			args[i]=self.constr.fixed[i].value
		ll=logl.LL(args, self.panel, constraints=self.constr,print_err=True)
		if ll.LL is None:
			if self.panel.options.loadargs.value:
				print("Initial arguments failed, attempting default OLS-arguments ...")
				self.panel.args.set_init_args(self.panel,default=True)
				ll=logl.LL(self.panel.args.args_OLS,self.panel,constraints=self.constr,print_err=True)
				if ll.LL is None:
					raise RuntimeError("OLS-arguments failed too, you should check the data")
				else:
					print("default OLS-arguments worked")
			else:
				raise RuntimeError("OLS-arguments failed, you should check the data")
				
		return ll
	
	def normalize(self,dx):
		args_v=self.ll.args.args_v
		dx_norm=(args_v!=0)*dx/(np.abs(args_v)+(args_v==0))
		dx_norm=(args_v<1e-2)*dx+(args_v>=1e-2)*dx_norm	
		return dx_norm	

def hessin(hessian):
	try:
		h=-np.linalg.inv(hessian)
	except:
		return None	
	return h

def nummerical_hessin(g,g_old,hessin,dxi):
	#Not currently in use
	if dxi is None:
		return None
	dg=g-g_old 				#Compute difference of gradients,
	#and difference times current matrix:
	n=len(g)
	hdg=(np.dot(hessin,dg.reshape(n,1))).flatten()
	fac=fae=sumdg=sumxi=0.0 							#Calculate dot products for the denominators. 
	fac = np.sum(dg*dxi) 
	fae = np.sum(dg*hdg)
	sumdg = np.sum(dg*dg) 
	sumxi = np.sum(dxi*dxi) 
	if (fac > (3.0e-16*sumdg*sumxi)**0.5):#Skip update if fac not sufficiently positive.
		fac=1.0/fac
		fad=1.0/fae 
								#The vector that makes BFGS different from DFP:
		dg=fac*dxi-fad*hdg   
		#The BFGS updating formula:
		hessin+=fac*dxi.reshape(n,1)*dxi.reshape(1,n)
		hessin-=fad*hdg.reshape(n,1)*hdg.reshape(1,n)
		hessin+=fae*dg.reshape(n,1)*dg.reshape(1,n)	
	return hessin


def solve(constr,H, g, x):
	"""Solves a second degree taylor expansion for the dc for df/dc=0 if f is quadratic, given gradient
	g, hessian H, inequalty constraints c and equalitiy constraints c_eq and returns the solution and 
	and index constrained indicating the constrained variables"""
	if H is None:
		raise RuntimeError('Hessian is None')
	if constr is None:
		return -np.linalg.solve(H,g).flatten()	
	n=len(H)
	k=len(constr)
	H=np.concatenate((H,np.zeros((n,k))),1)
	H=np.concatenate((H,np.zeros((k,n+k))),0)
	g=np.append(g,(k)*[0])
	
	for i in range(k):
		H[n+i,n+i]=1
	j=0
	dx=np.zeros(len(g))
	for i in constr.fixed:
		kuhn_tucker(constr.fixed[i],i,j,n, H, g, x,dx, recalc=False)
		j+=1
	dx=-np.linalg.solve(H,g).flatten()	
	OK=False
	w=0
	for r in range(50):
		j2=j
		
		for i in constr.intervals:
			dx=kuhn_tucker(constr.intervals[i],i,j2,n, H, g, x,dx)
			j2+=1
		OK=constr.within(x+dx[:n],False)
		if OK: 
			break
		if r==k+3:
			#print('Unable to set constraints in direction calculation')
			break

	return dx[:n]


def solve_delete(constr,H, g, x):
	"""Solves a second degree taylor expansion for the dc for df/dc=0 if f is quadratic, given gradient
	g, hessian H, inequalty constraints c and equalitiy constraints c_eq and returns the solution and 
	and index constrained indicating the constrained variables"""
	if H is None:
		return None,g*0
	try:
		list(constr.keys())[0]
	except:
		return -np.linalg.solve(H,g).flatten()	
	
	m=len(H)
	
	idx=np.ones(m,dtype=bool)
	delmap=np.arange(m)
	if len(list(constr.fixed.keys()))>0:#removing fixed constraints from the matrix
		idx[list(constr.fixed.keys())]=False
		H=H[idx][:,idx]
		g=g[idx]
		delmap-=np.cumsum(idx==False)
		delmap[idx==False]=m#if for some odd reason, the deleted variables are referenced later, an out-of-bounds error is thrown
	n=len(H)
	k=len(constr.intervals)
	H=np.concatenate((H,np.zeros((n,k))),1)
	H=np.concatenate((H,np.zeros((k,n+k))),0)
	g=np.append(g,(k)*[0])
	
	for i in range(k):
		H[n+i,n+i]=1
	dx=-np.linalg.solve(H,g).flatten()	
	xi_full=np.zeros(m)
	OK=False
	keys=list(constr.intervals.keys())
	for r in range(50):		
		for j in range(k):
			key=keys[j]
			dx=kuhn_tucker_del(constr,key,j,n, H, g, x,dx,delmap)
		xi_full[idx]=dx[:n]
		OK=constr.within(x+xi_full,False)
		if OK: 
			break
		if r==k+3:
			#print('Unable to set constraints in direction calculation')
			break
	xi_full=np.zeros(m)
	xi_full[idx]=dx[:n]
	return xi_full

def kuhn_tucker_del(constr,key,j,n,H,g,x,dx,delmap,recalc=True):
	q=None
	c=constr.intervals[key]
	i=delmap[key]
	if not c.value is None:
		q=-(c.value-x[i])
	elif x[i]+dx[i]<c.min:
		q=-(c.min-x[i])
	elif x[i]+dx[i]>c.max:
		q=-(c.max-x[i])
	if q!=None:
		H[i,n+j]=1
		H[n+j,i]=1
		H[n+j,n+j]=0
		g[n+j]=q
		if recalc:
			dx=-np.linalg.solve(H,g).flatten()	
	return dx


def kuhn_tucker(c,i,j,n,H,g,x,dx,recalc=True):
	q=None
	if not c.value is None:
		q=-(c.value-x[i])
	elif x[i]+dx[i]<c.min:
		q=-(c.min-x[i])
	elif x[i]+dx[i]>c.max:
		q=-(c.max-x[i])
	if q!=None:
		H[i,n+j]=1
		H[n+j,i]=1
		H[n+j,n+j]=0
		g[n+j]=q
		if recalc:
			dx=-np.linalg.solve(H,g).flatten()	
	return dx





#!/usr/bin/env python
# -*- coding: utf-8 -*-

#used for debugging
import numpy as np
import stat_functions as stat
import calculus_functions as cf
import time
import functions as fu
import os
import loglikelihood as lgl



def hess_debug(ll,panel,g,d):
	"""Calculate the Hessian nummerically, using the analytical gradient. For debugging. Assumes correct and debuggeed gradient"""
	x=ll.args.args_v
	n=len(x)
	dx=np.identity(n)*d
	H=np.zeros((n,n))
	ll0=lgl.LL(x,panel)
	f0=g.get(ll0)
	for i in range(n):
		ll=lgl.LL(x+dx[i],panel)
		if not ll is None:
			f1=g.get(ll)
			H[i]=(f1-f0)/d

			
	return H

def grad_debug(ll,panel,d):
	"""Calcualtes the gradient numerically. For debugging"""
	x=ll.args.args_v
	n=len(x)
	dx=np.abs(x.reshape(n,1))*d
	dx=dx+(dx==0)*d
	dx=np.identity(n)*dx

	g=np.zeros(n)
	f0=lgl.LL(x,panel)
	for i in range(n):
		for j in range(5):
			dxi=dx[i]*(0.5**j)
			f1=lgl.LL(x+dxi,panel)
			if not f1 is None:
				g[i]=(f1.LL-f0.LL)/dxi[i]
				break
	return g




def grad_debug_allparams(f0,panel,d,varname,pos=0):
	args=fu.copy_array_dict(f0.args.args_d)
	args[varname][pos]+=d
	f1=lgl.LL(args, panel)
	dLL=(f1.LL-f0.LL)/d
	
	for i in f1.__dict__.keys():
		x0=f0.__dict__[i]
		x1=f1.__dict__[i]
		if (type(x1)==np.ndarray):
			print(i)
			print((np.sum(x1-x0)/d))
	#LL_calc(f0, panel, d)
	a=0
	
	
def grad_debug_allparams_diff(f0,panel):

	f1=lgl.LL(f0.args.args_d, panel)
	for i in f1.__dict__.keys():
		x0=f0.__dict__[i]
		x1=f1.__dict__[i]
		if (type(x1)==np.ndarray):
			print(i)
			print((np.sum(x1-x0)))
	#LL_calc(f0, panel, d)
	a=0
	
def grad_debug_detail(f0,panel,d,llname,varname1,pos1=0):
	args1=fu.copy_array_dict(f0.args.args_d)
	args1[varname1][pos1]+=d
	
	f0=lgl.LL(f0.args.args_d, panel)
	f1=lgl.LL(args1, panel)

	if type(llname)==list or type(llname)==tuple:
		ddL=(f1.__dict__[llname[0]].__dict__[llname[1]]-f0.__dict__[llname[0]].__dict__[llname[1]])/d
	else:
		ddL=(f1.__dict__[llname]-f0.__dict__[llname])/d
	return ddL


def grad_debug_detail_oldll(f0,panel,d,llname,varname1,pos1=0):
	args1=fu.copy_array_dict(f0.args_d)
	args1[varname1][pos1]+=d
	

	f1=lgl.LL(args1, panel)

	if type(llname)==list:
		ddL=(f1.__dict__[llname[0]].__dict__[llname[1]]-f0.__dict__[llname[0]].__dict__[llname[1]])
	else:
		ddL=(f1.__dict__[llname]-f0.__dict__[llname])
	return ddL
	
	
def hess_debug_detail(f0,panel,d,llname,varname1,varname2,pos1=0,pos2=0):
	args1=fu.copy_array_dict(f0.args.args_d)
	args2=fu.copy_array_dict(f0.args.args_d)
	args3=fu.copy_array_dict(f0.args.args_d)
	args1[varname1][pos1]+=d
	args2[varname2][pos2]+=d	
	args3[varname1][pos1]+=d
	args3[varname2][pos2]+=d
	f1=lgl.LL(args1, panel)
	f2=lgl.LL(args2, panel)
	f3=lgl.LL(args3, panel)
	if type(llname)==list:
		ddL=(f3.__dict__[llname[0]].__dict__[llname[1]]-f2.__dict__[llname[0]].__dict__[llname[1]]
		     -f1.__dict__[llname[0]].__dict__[llname[1]]+f0.__dict__[llname[0]].__dict__[llname[1]])/(d**2)
	else:
		ddL=(f3.__dict__[llname]-f2.__dict__[llname]-f1.__dict__[llname]+f0.__dict__[llname])/(d**2)
	return ddL
	
	
def LL_calc2(ll,panel,d,X=None):
	self=ll
	args=self.args.args_d#using dictionary arguments
	args['beta'][3]+=d
	if X is None:
		X=panel.XIV
	matrices=lgl.set_garch_arch(panel,args)
	if matrices is None:
		return None		

	AMA_1,AMA_1AR,GAR_1,GAR_1MA=matrices
	(N,T,k)=X.shape

	u = panel.Y-cf.dot(X,args['beta'])
	e = cf.dot(AMA_1AR,u)
	lnv_ARMA = self.garch(panel, args, GAR_1MA,e)
	W_omega = cf.dot(panel.W_a,args['omega'])
	lnv = W_omega+lnv_ARMA# 'N x T x k' * 'k x 1' -> 'N x T x 1'
	grp = self.group_variance(panel, lnv, e,args)
	lnv+=grp
	lnv = np.maximum(np.minimum(lnv,100),-100)
	v = np.exp(lnv)*panel.a[3]
	v_inv = np.exp(-lnv)*panel.a[3]
	e_RE = self.re_obj.RE(e)
	return e_RE
	e_REsq = e_RE**2
	LL = self.LL_const-0.5*np.sum((lnv+(e_REsq)*v_inv)*panel.included[3])
	
	if abs(LL)>1e+100: 
		return None
	self.AMA_1,self.AMA_1AR,self.GAR_1,self.GAR_1MA=matrices
	self.u,self.e, self.lnv_ARMA        = u,         e,       lnv_ARMA
	self.lnv,self.v,self.v_inv          = lnv,       v,       v_inv
	self.e_RE,self.e_REsq               = e_RE,      e_REsq

	return LL

def LL_calc(ll,panel,d,X=None):
	self=ll
	args=self.args.args_d#using dictionary arguments
	if X is None:
		X=panel.XIV
	matrices=lgl.set_garch_arch(panel,args)
	if matrices is None:
		return None		

	AMA_1,AMA_1AR,GAR_1,GAR_1MA=matrices
	(N,T,k)=X.shape

	u = panel.Y-cf.dot(X,args['beta'])
	e = cf.dot(AMA_1AR,u)
	lnv_ARMA = self.garch(panel, args, GAR_1MA,e)
	W_omega = cf.dot(panel.W_a,args['omega'])
	lnv = W_omega+lnv_ARMA# 'N x T x k' * 'k x 1' -> 'N x T x 1'
	grp = self.group_variance(panel, lnv, e,args)
	lnv+=grp
	lnv = np.maximum(np.minimum(lnv,100),-100)
	v = np.exp(lnv)*panel.a[3]
	v_inv = np.exp(-lnv)*panel.a[3]	
	e_RE = self.re_obj.RE(e)+d*panel.included[3]
	e_REsq = e_RE**2
	return -0.5*((lnv+(e_REsq)*v_inv)*panel.included[3])
	LL = self.LL_const-0.5*np.sum((lnv+(e_REsq)*v_inv)*panel.included[3])
	
	if abs(LL)>1e+100: 
		return None
	self.AMA_1,self.AMA_1AR,self.GAR_1,self.GAR_1MA=matrices
	self.u,self.e, self.lnv_ARMA        = u,         e,       lnv_ARMA
	self.lnv,self.v,self.v_inv          = lnv,       v,       v_inv
	self.e_RE,self.e_REsq               = e_RE,      e_REsq

	return LL

def LL_calc_custom(ll,panel,d,X=None):
	f0=LL_calc(ll,panel,0,X=None)
	f1=LL_calc(ll,panel,d,X=None)
	dLLeRE=(f1-f0)/d
	
	f0=LL_calc2(ll,panel,0,X=None)
	f1=LL_calc2(ll,panel,d,X=None)
	deRE=(f1-f0)/d
	return dLLeRE,deRE

	

def LL_calc_debug(ll,panel,g,d):
	f0=LL_calc(ll, panel,0)
	f1=LL_calc(ll, panel,d)
	f2=LL_calc(ll, panel,d*2)
	d_x=[]
	for i in range(5):
		d_x.append((f1[i]-f0[i])/d)
		print (np.sum(d_x[i]))

	dLL1=np.sum(g.DLL_e*d_x[1])
	dLL2=np.sum(g.DLL_e*d_x[2])
	dLL3=np.sum(g.DLL_e*d_x[3])
	
	#dd=np.sum(f2[2]-2*f1[2]+f0[2])/(d**2)
	a=0
	
	
def pedantic(ll,msg,lmbda,ok,rev,direction,mp,its,incr,po,prev_dx,LL0,diff_log,ll0,dxLL):
	"""Used in maximize.maximize() when the LL maximization does not converge. It is usually 
	easier to identify errors in gradient or hessian close to the maximum."""
	ll,msg2,lmbda2,ok,rev=lnsrch(ll,direction,mp,its,incr,po,prev_dx,direction.dx,max_its=6)
	dx=direction.dx
	n=len(dx)
	dx_norm=direction.dx_norm
	h=np.diag(direction.H)
	g=direction.g
	
	if ll.panel.options.pedantic.value==1:
		return ll,msg,lmbda,ok,rev
	reversed_direction=g*dx<0

	for sgn in [1,-1]:
		if ll.LL-LL0<1e-6*len(dx) and sum(reversed_direction)>0:

			max_arg=((dxLL)==max((dxLL[reversed_direction])))
			for sel,d,s in [
				(max_arg==False,			  dx, 'directions other than the maxiumum'),
				(max_arg==True,				  dx, 'the maxiumum'),				
				(reversed_direction,	      dx, 'reversed directions'),
				(reversed_direction==False,	  dx, 'directions other than the reversed and reversing the reversed directions')
				]:
				dxi=sgn*np.array(d)
				dxi[sel]=0
				ll2,msg,lmbda2,ok,rev=lnsrch(ll,direction,mp,its,incr,po,prev_dx,dxi,max_its=6,convex_action='ignore')
				if ll2.LL>ll.LL:
					if not printout_func(1.0,f"Improved {ll2.LL-ll.LL} after fixing {s}",ll,its,direction,incr,po,1,task='linesearch'):return ll,direction,po
					lmbda+=lmbda2
					ll=ll2
					break

	if ll.LL-LL0<1e-6*len(dx):
		if len(diff_log)>2:
			avg_dir=np.mean(np.array(diff_log)[-8:],0)
			ll2,msg,lmbda2,ok,rev=lnsrch(ll,direction,mp,its,incr,po,prev_dx,avg_dir,max_its=6,convex_action='ignore')
			if ll2.LL>ll.LL:
				if not printout_func(1.0,f"Improved {ll2.LL-ll.LL} after following the average past directions",ll,its,direction,incr,po,1,task='linesearch'):return ll,direction,po
				ll=ll2
				lmbda+=lmbda2
	if ll.LL-LL0==0 and ll.panel.options.pedantic.value==3:
		if not printout_func(1.0,f"Doing brute force",ll,its,direction,incr,po,1,task='linesearch'):return ll,direction,po
		ll,msg,lmbda2,ok,rev=brute_force(ll, mp, direction, its, incr, po, prev_dx, msg, lmbda, ok, 4)		
		lmbda+=lmbda2		
	diff_log.append(ll.args.args_v-ll0.args.args_v)
	
	return ll,msg,lmbda,ok,rev


		


def brute_force(ll,mp,direction,its,incr,po,prev_dx,msg,lmbda,ok,max_its):
	"""Brute force maximization for pedantic(), when everything else fails."""
	dx=direction.dx
	rng=np.arange(len(dx))
	for i in rng:
		for sign in [-1,1]:
			if not (i in direction.constr.fixed):
				ll2,msg,lmbda,ok,rev=lnsrch(ll,direction,mp,its,incr,po,prev_dx,sign*dx*(rng==i),max_its,convex_action='ignore')
				if ll2.LL>ll.LL:
					ll=ll2
					break
	return ll,msg,lmbda,ok,rev


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import stat_functions as stat
import functions as fu
import calculus_functions as cf

MAX_COLLINEARITY=1e+7
EXTREEME_COLLINEARITY=1E+20
SMALL_COLLINEARITY=30



class constraint:
	def __init__(self,index,assco,cause,value, interval,names,category):
		self.name=names[index]
		self.intervalbound=None
		self.max=None
		self.min=None
		self.value=None
		self.value_str=None
		if interval is None:
			self.value=value
			self.value_str=str(round(self.value,8))
		else:
			if interval[0]>interval[1]:
				raise RuntimeError('Lower constraint cannot exceed upper')
			self.min=interval[0]
			self.max=interval[1]
			self.cause='user/general constraint'
		self.assco_ix=assco
		if assco is None:
			self.assco_name=None
		else:
			self.assco_name=names[assco]
		self.cause=cause
		self.category=category	
		
class constraints(dict):

	"""Stores the constraints of the LL maximization"""	
	def __init__(self,direction,args,its=0):
		dict.__init__(self)
		self.categories={}
		self.fixed={}
		self.intervals={}
		self.associates={}
		self.collinears={}
		self.args=args
		self.panel_args=direction.panel.args
		self.CI=None
		self.its=its
		self.pqdkm=direction.panel.pqdkm
		self.m_zero=direction.panel.m_zero

	def add(self,name,assco,cause,interval=None,replace=True,value=None,args=None):
		#(self,index,assco,cause,interval=None,replace=True,value=None)
		name,index=self.panel_args.get_name_ix(name)
		name_assco,assco=self.panel_args.get_name_ix(assco,True)
		for i in index:
			self.add_item(i,assco,cause, interval ,replace,value,args)
		
	def clear(self,cause=None):
		for c in list(self.keys()):
			if self[c].cause==cause or cause is None:
				self.delete(c)	

	def add_item(self,index,assco,cause,interval,replace,value,args):
		"""Adds a constraint. 'index' is the position
		for which the constraints shall apply.  \n\n

		Equality constraints are chosen by specifying 'minimum_or_value' \n\n
		Inequality constraints are chosen specifiying 'maximum' and 'minimum'\n\n
		'replace' determines whether an existing constraint shall be replaced or not 
		(only one equality and inequality allowed per position)"""
		
		if args is None:
			args=self.panel_args

		if not replace:
			if index in self:
				return False
		interval,value=test_interval(interval, value)
		if interval is None:#this is a fixed constraint
			if len(self.fixed)==len(args.names_v)-1:#can't lock all variables
				return False
			if value is None:
				value=self.args.args_v[index]
			if index in self.intervals:
				c=self[index]
				if not (c.min<value<c.max):
					return False
				else:
					self.intervals.pop(index)
		elif index in self.fixed: #this is an interval constraint, no longer a fixed constraint
			self.fixed.pop(index)

		eq,category,j=self.panel_args.positions_map[index]
		if not category in self.categories:
			self.categories[category]=[index]
		elif not index in self.categories[category]:
			self.categories[category].append(index)

		c=constraint(index,assco,cause,value, interval ,args.names_v,category)
		self[index]=c
		if value is None:
			self.intervals[index]=c
		else:
			self.fixed[index]=c
		if not assco is None:
			if not assco in self.associates:
				self.associates[assco]=[index]
			elif not index in self.associates[assco]:
				self.associates[assco].append(index)
		if cause=='collinear':
			self.collinears[index]=assco
		return True
		
	def delete(self,index):
		if not index in self:
			return False
		self.pop(index)
		if index in self.intervals:
			self.intervals.pop(index)
		if index in self.fixed:
			self.fixed.pop(index)		
		eq,category,j=self.panel_args.positions_map[index]
		c=self.categories[category]
		if len(c)==1:
			self.categories.pop(category)
		else:
			i=np.nonzero(np.array(c)==index)[0][0]
			c.pop(i)
		a=self.associates
		for i in a:
			if index in a[i]:
				if len(a[i])==1:
					a.pop(i)
					break
				else:
					j=np.nonzero(np.array(a[i])==index)[0][0]
					a[i].pop(j)
		if index in self.collinears:
			self.collinears.pop(index)
		return True
		

	def set_fixed(self,x):
		"""Sets all elements of x that has fixed constraints to the constraint value"""
		for i in self.fixed:
			x[i]=self.fixed[i].value
		
	def within(self,x,fix=False):
		"""Chekcs if x is within interval constraints. If fix=True, then elements of
		x outside constraints are set to the nearest constraint. if fix=False, the function 
		returns False if x is within constraints and True otherwise"""
		for i in self.intervals:
			c=self.intervals[i]
			if (c.min<=x[i]<=c.max):
				c.intervalbound=None
			else:
				if fix:
					x[i]=max((min((x[i],c.max)),c.min))
					c.intervalbound=str(round(x[i],8))
				else:
					return False
		return True
	
	def add_static_constraints(self,overflow_problem):
		pargs=self.panel_args
		p, q, d, k, m=self.pqdkm
		
		self.add_custom_constraints(pargs.user_constraints)
		if overflow_problem:
			general_constraints=[('rho',-0.5,0.5),('lambda',-0.5,0.5),('gamma',-0.5,0.5),('psi',-0.5,0.5)]
			self.add_custom_constraints(general_constraints)
		else:
			general_constraints=[('rho',-2,2),('lambda',-2,2),('gamma',-2,2),('psi',-2,2)]
			self.add_custom_constraints(general_constraints)
			
		if self.its==0 and p>0:
			if pargs.args_init.args_d['rho'][0]==0:
				self.add(pargs.positions['rho'][0],None,'Initial MA constr',value=0.0)
		if self.m_zero:
			self.add(pargs.positions['psi'][0],None,'GARCH input constr',value=0.05)
			
		if k*m==0:
			return
		if sum(self.args.args_v[pargs.positions['psi']]**2)==0:
			for i in pargs.positions['gamma']:
				self.add(i,None,'GARCH term cannot be positive if ARCH terms are zero',value=0)
	
	def add_dynamic_constraints(self,direction):
		ll=direction.ll
		k,k=direction.H.shape
		self.weak_mc_dict=dict()
		include=np.array(k*[True])
		include[list(direction.constr.fixed)]=False
		self.H_correl_problem=constraint_correl_cluster(direction,include)
		self.mc_problems=[]#list of [index,associate,condition index]
		self.CI=constraint_multicoll(k, direction, include, self.mc_problems)
		add_mc_constraint(direction,self.mc_problems,self.weak_mc_dict)
		select_arma(direction.constr, ll)
	
	def add_custom_constraints(self,constraints,replace=True,cause='user constraint',clear=False,args=None):
		"""Adds custom range constraints\n\n
			constraints shall be on the format [(name, minimum, maximum), ...]
			or
			a dictionary of the same form of an argument dictionary
			(a dictionary of dictonaries on the form dict[category][name]) 
			where items are lists of [minimum,maximum] or the fixing value
			
			name is taken from self.panel_args.names_v"""	
		
		if clear:
			self.clear(cause)
		if type(constraints)==list or type(constraints)==tuple:
			for c in constraints:
				self.add_custom_constraints_list(c,replace,cause,args)
		elif type(constraints)==dict:
			self.add_custom_constraints_dict(constraints,replace,cause,args)
		else:
			raise TypeError("The constraints must be a list, tuple or dict.")
	
	
	def add_custom_constraints_list(self,constraint,replace,cause,args):
		"""Adds a custom range constraint\n\n
			constraint shall be on the format (name, minimum, maximum)"""
		if type(constraint)==str or isinstance(constraint,int):
			name, value=constraint,None
			self.add(name,None,cause,replace=replace,value=value,args=args)
			return
		elif len(constraint)==2:
			name, minimum,maximum=list(constraint)+[None]
		elif len(constraint)==3:
			name, minimum, maximum=constraint
		else:
			raise TypeError("A constraint needs to be a string, integer or iterable with lenght no more than tree.")

		name,ix=self.panel_args.get_name_ix(name)
		m=[minimum,maximum]
		for i in range(2):
			if type(m[i])==str:
				try:
					m[i]=eval(m[i],globals(),self.direction.ll.__dict__)
				except:
					print(f"Custom constraint {name} ({m[i]}) failed")
					return
		[minimum,maximum]=m
		self.add(name,None,cause, [minimum,maximum],replace,args=args)
		
	def add_custom_constraints_dict(self,constraints,replace,cause,args):
		"""Adds a custom range constraint\n\n
		   If list, constraint shall be on the format (minimum, maximum)"""
		for grp in constraints:
			for name in constraints[grp]:	
				c=constraints[grp][name]
				if type(c)==list:
					self.add(name,None,cause, c,replace,args=args)
				else:
					self.add(name,None,cause, [c,None],replace,args=args)
	
	
def test_interval(interval,value):
	if not interval is None:
		if np.any([i is None for i in interval]):
			if interval[0] is None:
				value=interval[1]
			else:
				value=interval[0]
			interval=None	
	return interval,value

def append_to_ID(ID,intlist):
	inID=False
	for i in intlist:
		if i in ID:
			inID=True
			break
	if inID:
		for j in intlist:
			if not j in ID:
				ID.append(j)
		return True
	else:
		return False

def correl_IDs(p):
	IDs=[]
	appended=False
	x=np.array(p[:,1:3],dtype=int)
	for i,j in x:
		for k in range(len(IDs)):
			appended=append_to_ID(IDs[k],[i,j])
			if appended:
				break
		if not appended:
			IDs.append([i,j])
	g=len(IDs)
	keep=g*[True]
	for k in range(g):
		if keep[k]:
			for h in range(k+1,len(IDs)):
				appended=False
				for m in range(len(IDs[h])):
					if IDs[h][m] in IDs[k]:
						appended=append_to_ID(IDs[k],  IDs[h])
						keep[h]=False
						break
	g=[]
	for i in range(len(IDs)):
		if keep[i]:
			g.append(IDs[i])
	return g

def normalize(H,include):
	C=-H[include][:,include]
	d=np.maximum(np.diag(C).reshape((len(C),1)),1e-30)**0.5
	C=C/(d*d.T)
	includemap=np.arange(len(include))[include]
	return C,includemap
	
def decomposition(H,include=None):
	if include is None:
		include=[True]*len(H)
	C,includemap=normalize(H, include)
	c_index,var_prop=stat.var_decomposition(XXNorm=C)
	c_index=c_index.flatten()
	return c_index, var_prop,includemap
	
	
def multicoll_problems(direction,H,include,mc_problems):
	c_index, var_prop, includemap = decomposition(H, include)
	if c_index is None:
		return False,False
	mc_list=[]
	largest_ci=None
	for cix in range(1,len(c_index)):
		if ((np.sum(var_prop[-cix]>0.5)>1) or ((np.sum(var_prop[-cix]>0.3)>1) and (np.sum(var_prop[-cix]>0.99)>0))) and (c_index[-cix]>SMALL_COLLINEARITY):
			if largest_ci is None:
				largest_ci=c_index[-cix]
			var_prop_ix=np.argsort(var_prop[-cix])[::-1]
			var_prop_val=var_prop[-cix][var_prop_ix]
			j=var_prop_ix[0]
			j=includemap[j]
			done=var_prop_check(direction,var_prop_ix, var_prop_val, includemap,j,mc_problems,c_index[-cix],mc_list)
			if done:
				break
	if len(mc_list)==0:
		return  c_index[-1],mc_list
	return c_index[-1],mc_list

def var_prop_check(direction,var_prop_ix,var_prop_val,includemap,assc,mc_problems,cond_index,mc_list):
	if cond_index>EXTREEME_COLLINEARITY:
		lim=0.3
	else:
		lim=0.5
	for i in range(1,len(var_prop_ix)):
		if var_prop_val[i]<lim:
			return True
		index=var_prop_ix[i]
		index=includemap[index]
		mc_problems.append([index,assc,cond_index])
		mc_list.append(index)
		return False
		
def add_mc_constraint(direction,mc_problems,weak_mc_dict):
	"""Adds constraints for severe MC problems"""
	constr=direction.constr
	if len(mc_problems)==0:
		return
	no_check=get_no_check(direction)
	a=[i[0] for i in mc_problems]
	if no_check in a:
		mc=mc_problems[a.index(no_check)]
		mc[0],mc[1]=mc[1],mc[0]
	mc_limit=MAX_COLLINEARITY
	if direction.increment==0:
		mc_limit=SMALL_COLLINEARITY
	for index,assc,cond_index in mc_problems:
		if (not index in weak_mc_dict) and (cond_index>SMALL_COLLINEARITY) and (cond_index<=mc_limit) and (not index==no_check):
			weak_mc_dict[index]=[assc,cond_index]#contains also collinear variables that are only slightly collinear, which shall be restricted when calcuating CV-matrix.	
		if not ((index in constr.associates) or (index in constr.collinears)) and cond_index>mc_limit and (not index==no_check):
			constr.add(index,assc,'collinear')
		
def get_no_check(direction):
	no_check=direction.panel.options.do_not_constrain.value
	X_names=direction.panel.input.X_names
	if not no_check is None:
		if no_check in X_names:
			return X_names.index(no_check)
		print("A variable was set for the 'Do not constraint' option (do_not_constrain), but it is not among the x-variables")
			
	
		


def constraint_multicoll(k,direction,include,mc_problems):
	CI_max=0
	for i in range(k-1):
		CI,mc_list=multicoll_problems(direction,direction.H,include,mc_problems)
		CI_max=max((CI_max,CI))
		if len(mc_list)==0:
			break
		include[mc_list]=False
	return CI_max
		
def constraint_correl_cluster(direction,include):
	if int(direction.its/2)==direction.its/2 and direction.its>0:
		return False		
	H=direction.H
	dH=np.diag(H).reshape((len(H),1))
	corr=np.abs(H/(np.abs(dH*dH.T)+1e-100)**0.5)
	np.fill_diagonal(corr,0)
	problems=list(np.unique(np.nonzero((corr>1-1e-12)*(corr<1))[0]))
	if len(problems)==0:
		return False
	#not_problem=np.argsort(dH[problems].flatten())[0]
	#problems.pop(not_problem)
	allix=list(range(len(H)))
	for i in problems:
		allix.pop(allix.index(i))
	if len(allix)==0:
		return	False
	not_problem=np.argsort(dH[allix].flatten())[-1]
	allix.pop(not_problem)	
	for i in allix:
		include[i]=False
		if not i in direction.constr.fixed:
			direction.constr.add(i,None,'Perfect correlation in hessian')
	return True
		

	

def select_arma(constr,ll):
	for i in constr.collinears:
		if constr[i].category in ['rho','lambda', 'gamma','psi']:
			assc=constr.collinears[i]
			if not assc in constr.fixed:
				if abs(ll.args.args_v[assc])<abs(ll.args.args_v[i]):
					constr.delete(i)
					constr.add(assc,i,'collinear')
	
	
			
		






#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import functions as fu

class re_obj:
	def __init__(self,panel,group,T_i,T_i_count,fixed_random_eff):
		"""Following Greene(2012) p. 413-414"""
		if fixed_random_eff==0:
			self.panel=panel
			self.FE_RE=0
			return
		self.panel=panel
		self.sigma_u=0
		self.group=group
		self.avg_Tinv=np.mean(1/T_i_count) #T_i_count is the total number of observations for each group (N,1)
		self.T_i=T_i*panel.included[3]#self.T_i is T_i_count at each observation (N,T,1)
		self.FE_RE=fixed_random_eff
	
	def RE(self,x,recalc=True):
		if self.FE_RE==0:
			return np.zeros(x.shape)
		panel=self.panel
		if self.FE_RE==1:
			self.xFE=self.FRE(x)
			return self.xFE
		if recalc:
			N,T,k=x.shape
			
			self.xFE=(x+self.FRE(x))*panel.included[3]
			self.e_var=self.panel.mean(self.xFE**2)/(1-self.avg_Tinv)
			self.v_var=self.panel.mean((panel.included[3]*x)**2)-self.e_var
			if self.v_var<0:
				#print("Warning, negative group random effect variance. 0 is assumed")
				self.v_var=0
				self.theta=panel.zeros[3]
				return panel.zeros[3]
			self.theta=(1-np.sqrt(self.e_var/(self.e_var+self.v_var*self.T_i)))*self.panel.included[3]
			self.theta*=(self.T_i>1)
			if np.any(self.theta>1) or np.any(self.theta<0):
				raise RuntimeError("WTF")
		#x=panel.Y
		eRE=self.FRE(x,self.theta)
		return eRE
	
	def dRE(self,dx,x,vname):
		"""Returns the first and second derivative of RE"""
		if dx is None:
			return None
		if self.FE_RE==0:
			return np.zeros(dx.shape)		
		panel=self.panel
		if not hasattr(self,'dxFE'):
			self.dxFE=dict()
			self.dFE_var=dict()
			self.dtheta=dict()
			self.de_var=dict()
			self.dv_var=dict()
		
		if dx is None:
			return None
		elif self.FE_RE==1:
			return self.FRE(dx)	
		if self.v_var==0:
			return np.zeros(dx.shape)
		(N,T,k)=dx.shape	

		self.dxFE[vname]=(dx+self.FRE(dx))*panel.included[3]
		self.de_var[vname]=2*np.sum(np.sum(self.xFE*self.dxFE[vname],0),0)/(self.panel.NT*(1-self.avg_Tinv))
		self.dv_var[vname]=(2*np.sum(np.sum(x*dx*panel.included[3],0),0)/self.panel.NT)-self.de_var[vname]		

		self.dtheta_de_var=(-0.5*(1/self.e_var)*(1-self.theta)*self.theta*(2-self.theta))
		self.dtheta_dv_var=(0.5*(self.T_i/self.e_var)*(1-self.theta)**3)
		self.dtheta[vname]=(self.dtheta_de_var*self.de_var[vname]+self.dtheta_dv_var*self.dv_var[vname])
		self.dtheta[vname]*=(self.T_i>1)
		dRE0=self.FRE(dx,self.theta)
		dRE1=self.FRE(x,self.dtheta[vname])
		ret=(dRE0+dRE1)*self.panel.included[3]
		remove_extreemes(ret)
		return ret
	
	def ddRE(self,ddx,dx1,dx2,x,vname1,vname2):
		"""Returns the first and second derivative of RE"""
		panel=self.panel
		if self.FE_RE==0:
			return 0*panel.included[4]		
		if dx1 is None or dx2 is None:
			return None
		(N,T,k)=dx1.shape
		(N,T,m)=dx2.shape			
		if self.sigma_u<0:
			return 0*panel.included[4]
		elif self.FE_RE==1:
			return self.FRE(ddx)	
		if self.v_var==0:
			return panel.zeros[4]

		if ddx is None:
			ddxFE=0
			ddx=0
			hasdd=False
		else:
			ddxFE=(ddx+self.FRE(ddx))*panel.included[4]
			hasdd=True
			
		dxFE1=self.dxFE[vname1].reshape(N,T,k,1)
		dxFE2=self.dxFE[vname2].reshape(N,T,1,m)
		dx1=dx1.reshape(N,T,k,1)
		dx2=dx2.reshape(N,T,1,m)
		de_var1=self.de_var[vname1].reshape(k,1)
		de_var2=self.de_var[vname2].reshape(1,m)
		dv_var1=self.dv_var[vname1].reshape(k,1)
		dv_var2=self.dv_var[vname2].reshape(1,m)		
		dtheta_de_var=self.dtheta_de_var.reshape(N,T,1,1)
		dtheta_dv_var=self.dtheta_dv_var.reshape(N,T,1,1)
		theta=self.theta.reshape(N,T,1,1)
		x=x.reshape(N,T,1,1)
		incl=panel.included[4]
		
		overflow=False
		theta_args=dxFE1, dxFE2, ddxFE, dx1, dx2, x, ddx, dtheta_dv_var, theta, dtheta_de_var, de_var1, de_var2, dv_var1, dv_var2
		try:
			ddtheta=self.calc_theta(*theta_args)
		except (RuntimeWarning,OverflowError) as e:
			print(e)
			remove_extreemes(theta_args)
			ddtheta=self.calc_theta(*theta_args)
			overflow=True
			
		if hasdd:
			dRE00=self.FRE(ddx,self.theta.reshape(N,T,1,1))
		else:
			dRE00=0
		dRE01=self.FRE(dx1,self.dtheta[vname2].reshape(N,T,1,m),True)
		dRE10=self.FRE(dx2,self.dtheta[vname1].reshape(N,T,k,1),True)
		dRE11=self.FRE(x,ddtheta,True)
		ret=(dRE00+dRE01+dRE10+dRE11)*panel.included[4]
		if overflow:
			remove_extreemes([ret])
		return (dRE00+dRE01+dRE10+dRE11)*panel.included[4]
	
	def calc_theta(self,dxFE1,dxFE2,ddxFE,dx1,dx2,x,ddx,dtheta_dv_var,theta,dtheta_de_var,de_var1,de_var2,dv_var1,dv_var2):
		incl=self.panel.included[4]
		(N,T,k,_)=dx1.shape
		(N,T,_,m)=dx2.shape		
		T_i=self.T_i.reshape(N,T,1,1)
		
		d2e_var=2*np.sum(np.sum(dxFE1*dxFE2+self.xFE.reshape(N,T,1,1)*ddxFE,0),0)/(self.panel.NT*(1-self.avg_Tinv))
		d2v_var=(2*np.sum(np.sum((dx1*dx2+x*ddx)*incl,0),0)/self.panel.NT)-d2e_var	
		
		d2theta_d_e_v_var=-0.5*dtheta_dv_var*(1/self.e_var)*(3*(theta-2)*theta+2)
		d2theta_d_v_var =-0.75*(T_i/self.e_var)**2*(1-theta)**5
		d2theta_d_e_var =-0.5*dtheta_de_var*(1/self.e_var)*(4-3*(2-theta)*theta)	
		
		ddtheta  =d2theta_d_e_var  * de_var1* de_var2 
		ddtheta +=d2theta_d_e_v_var * (de_var1* dv_var2+dv_var1* de_var2)
		ddtheta +=d2theta_d_v_var * dv_var1* dv_var2  
		ddtheta +=dtheta_de_var*d2e_var+dtheta_dv_var*d2v_var
		ddtheta*=(T_i>1)	
		
		return ddtheta
		
	def FRE(self,x,w=1,d=False):
		if self.group:
			return self.FRE_group(x,w,d)
		else:
			return self.FRE_time(x,w,d)
	
	def FRE_group(self,x,w,d):
		"""returns x after fixed effects, and set lost observations to zero"""
		#assumes x is a "N x T x k" matrix
		if x is None:
			return None
		T_i,s=get_subshapes(self.panel,x,True)
		incl=self.panel.included[len(s)]
		
		sum_x=np.sum(x*incl,1).reshape(s)
		mean_x=sum_x/T_i
		mean_x_all=np.sum(sum_x,0)/self.panel.NT
		try:
			dFE=w*(mean_x_all-mean_x)*incl#last product expands the T vector to a TxN matrix
		except (RuntimeWarning,OverflowError) as e:
			print(e)
			remove_extreemes([w])
			dFE=w*(mean_x_all-mean_x)*incl#last product expands the T vector to a TxN matrix
			remove_extreemes([dFE])
		return dFE
	
	def FRE_time(self,x,w,d):
		#assumes x is a "N x T x k" matrix
		
		
		if x is None:
			return None
		mean_x,mean_x_all,incl=mean_time(self.panel, x)
		try:
			dFE=(w*(mean_x_all-mean_x))*incl#last product expands the T vector to a TxN matrix
		except (RuntimeWarning,OverflowError) as e:
			print(e)
			remove_extreemes([w])
			dFE=(w*(mean_x_all-mean_x))*incl#last product expands the T vector to a TxN matrix		
			remove_extreemes([dFE])
		return dFE
	
def mean_time(panel,x,mean_dates=False):
	n_dates=panel.n_dates
	dmap=panel.date_map
	date_count,s=get_subshapes(panel,x,False)
	incl=panel.included[len(s)]
	x=x*incl
	sum_x_dates=np.zeros(s)
	for i in range(n_dates):
		sum_x_dates[i]=np.sum(x[dmap[i]],0)		
	mean_x_dates=sum_x_dates/date_count
	if mean_dates:
		return mean_x_dates
	mean_x=np.zeros(x.shape)
	for i in range(n_dates):
		mean_x[dmap[i]]=mean_x_dates[i]	
	mean_x_all=np.sum(sum_x_dates,0)/panel.NT
	return mean_x,mean_x_all,incl
	


def get_subshapes(panel,x,group):
	if group:
		if len(x.shape)==3:
			N,T,k=x.shape
			s=(N,1,k)
			T_i=panel.T_i
		elif len(x.shape)==4:
			N,T,k,m=x.shape
			s=(N,1,k,m)
			T_i=panel.T_i.reshape((N,1,1,1))	
		return T_i,s
	else:
		date_count=panel.date_count
		n_dates=panel.n_dates
		if len(x.shape)==3:
			N,T,k=x.shape
			s=(n_dates,1,k)
			
		elif len(x.shape)==4:
			N,T,k,m=x.shape
			s=(n_dates,1,k,m)
			date_count=date_count.reshape((n_dates,1,1,1))	
		return date_count,s			
	
	
	
def remove_extreemes(args,max_arg=1e+100):
	for i in range(len(args)):
		s=np.sign(args[i][np.abs(args[i])>max_arg])
		args[i][np.abs(args[i])>max_arg]=s*max_arg